"""General use utilitary methods"""

import inspect
import itertools
from pathlib import Path
from typing import Any, Iterable, List, Union

import numpy as np
from nexusformat.nexus.tree import NXfield

import assonant.data_classes.components as components
import assonant.data_classes.data_handlers as data_handlers
from assonant.data_classes import AssonantDataClass
from assonant.data_classes.components import Component
from assonant.data_classes.data_handlers import Axis, DataField, DataHandler, TimeSeries
from assonant.data_classes.enums import TransformationType


def get_component_classes():
    """Collect all classes in assonant.data_classes.components that subclass AssonantDataClass. Component
    class is an exception, as it is just a base class for components.
    """
    component_classes = []
    for name, obj in inspect.getmembers(components, inspect.isclass):
        if issubclass(obj, AssonantDataClass) and obj not in [AssonantDataClass, Component]:
            component_classes.append(obj)
    return component_classes


def get_data_handler_classes():
    """Collect all classes in assonant.data_classes.data_handlers that subclass AssonantDataClass. External
    class is an exception, as it is just a base class for components.
    """
    data_handler_classes = []
    for name, obj in inspect.getmembers(data_handlers, inspect.isclass):
        if issubclass(obj, AssonantDataClass) and obj not in [AssonantDataClass, DataHandler]:
            data_handler_classes.append(obj)
    return data_handler_classes


def value_equal_safe_agains_none_and_nan(v1: Union[int, float, str, bool], v2: Union[int, float, str, bool]) -> bool:
    """Compare two values if they are equal in a safe way against None and np.nan values.

    This method was create due to np.array_equals breaking if None value is passed when its equal_nan
    flag is set to True as np.isnan() methos used internally doesn't support None as input.

    Args:
        v1 (Union[int, float, str, bool]): Value to be compared agains v2.
        v2 (Union[int, float, str, bool]): Value to be compared agains v1.

    Returns:
        bool: True if v1 equals to v2 and False if not.
    """
    if type(v1) is not type(v2):
        # Cover comparison between None and np.nan. ALso guarantee this will not happen
        # on underlying comparisons
        return False
    elif v1 is not None and v2 is not None:
        # Cover comparison using np.nan.
        try:
            if not np.array_equal(v1, v2, equal_nan=True):
                return False
        except Exception:
            # Type may not be supported by operation with nan type, also meaning both values are not nan
            # Just compare them
            if not np.array_equal(v1, v2):
                return False
    else:
        # Cover comparison using None.
        if not np.array_equal(v1, v2):
            return False

    # Values were verified and no difference was found, that means v1 == v2
    return True


def array_equal_safe_against_none_and_nan(
    a1: Union[List, np.typing.NDArray], a2: Union[List, np.typing.NDArray]
) -> bool:
    """Compare two arrays/lists if they are element-wise equal in a safe way against None and np.nan values.

    This method was create due to np.array_equals breaking if None value is passed when its equal_nan
    flag is set to True as np.isnan() methos used internally doesn't support None as input.

    Args:
        a1 (Union[List, np.typing.NDArray]): Array or List to be compared against a2.
        a2 (Union[List, np.typing.NDArray]): Array or List to be compared against a1.

    Returns:
        bool: True if a1 equals to a2 and False if not.
    """
    if isinstance(a1, Iterable) and isinstance(a2, Iterable):
        if len(a1) != len(a2):
            # Different lengths, they are directly different.
            return False

        if np.array_equal(a1, a2):
            # Just a normal case of two equal arrays/lists.
            return True
        else:
            # Failure may be due to np.nan != np.nan so a specific comparsion need to be done
            # However, numpy method that does that breaks if None pass is passed. Due to that, an
            # element-wise comparison must be done and checking directly how to deal with specific cases.
            for v1, v2 in zip(a1, a2):
                if not value_equal_safe_agains_none_and_nan(v1, v2):
                    return False

            # All elements were verified and no difference was found, that means a1 == a2.
            return True
    else:
        if not value_equal_safe_agains_none_and_nan(a1, a2):
            return False

        # Values were verified and no difference was found, that means a1 == a2
        return True


def compare_dicts(d1, d2):
    """Recursively compare two dictionaries, handling numpy arrays."""

    if d1 is not None and d2 is not None:
        if d1.keys() != d2.keys():
            # If dict keys are different, they are clearly different
            return False

        for key in d1:
            # Check equality of each dict element
            v1 = d1[key]
            v2 = d2[key]

            if isinstance(v1, dict) and isinstance(v2, dict):
                if not compare_dicts(v1, v2):
                    return False
            elif not array_equal_safe_against_none_and_nan(v1, v2):
                return False

    return True


def get_fixture_values(fixture: Union[str, None], request) -> Union[Any, None]:
    """Wrapper class to handle getting fixture values based in their names.

    Args:
        fixture (Union[str, None]): Fixture name or None value.
        request (pytest.FixtureRequest): request fixture object from pytest.

    Returns:
       Union[Any, None]: Retrieved fixture value if not None else returns None.
    """

    fixture_value = request.getfixturevalue(fixture) if fixture is not None else None

    return fixture_value


def assert_data_field_content(data_field: DataField, input_value: Any, input_unit: Any, input_metadata: Any):
    """Assert that DataField was correctly created based on input values.

    Args:
        data_field (DataField): DataField which assertion will be applied over.
        input_value (Any): Input value parameter during DataField creation.
        input_unit (Any): Input unit parameter during DataField creation.
        input_metadata (Any): Input metadata parameter during DataField creation.

    Raise:
        AssertionError: Raises if any assertion fails.
    """
    assert isinstance(data_field, DataField)
    assert array_equal_safe_against_none_and_nan(data_field.value, input_value)
    assert value_equal_safe_agains_none_and_nan(data_field.unit, input_unit)
    assert compare_dicts(data_field.extra_metadata, input_metadata)


def assert_timeseries_content(
    timeseries: TimeSeries,
    input_value: Any,
    input_unit: Any,
    input_metadata: Any,
    input_timestamps: Any,
    input_timestamps_unit: Any,
    input_timestamps_metadata: Any,
):
    """Assert that TimeSeries was correctly created based on input values.

    Args:
        timeseries (TimeSeries): TimeSeries which assertion will be applied over.
        input_value (Any): Input value parameter during TimeSeries creation.
        input_unit (Any): Input unit parameter during TimeSeries creation.
        input_metadata (Any): Input metadata parameter during TimeSeries creation.
        input_timestamps (Any): Input timestamps parameter during TimeSeries creation.
        input_timestamps_unit (Any): Input timestamps unit parameter during TimeSeries creation.
        input_timestamps_metadata (Any): Input timestamps metadata parameter during TimeSeries creation.

    Raise:
        AssertionError: Raises if any assertion fails.
    """
    assert isinstance(timeseries, TimeSeries)
    assert_data_field_content(timeseries.value, input_value, input_unit, input_metadata)
    assert_data_field_content(timeseries.timestamps, input_timestamps, input_timestamps_unit, input_timestamps_metadata)


def assert_axis_content(
    axis: Axis,
    input_value: Any,
    input_unit: Any,
    input_metadata: Any,
    input_transformation_type: TransformationType,
    input_timestamps: Any = None,
    input_timestamps_unit: Any = None,
    input_timestamps_metadata: Any = None,
):
    """Assert that Axis was correctly created based on input values.

    Args:
        axis (Axis): Axis which assertion will be applied over.
        input_value (Any): Input value parameter during Axis creation.
        input_unit (Any): Input unit parameter during Axis creation.
        input_metadata (Any): Input metadata parameter during Axis creation.
        input_transformation_type(TransformationType): Input transformation type parameter during Axis creation.
        input_timestamps (Optional, Any): Input timestamps param during Axis creation. Defaults to None.
        input_timestamps_unit (Optional, Any): Input timestamps unit param during Axis creation. Defaults to None.
        input_timestamps_metadata (Optional, Any): Input timestamps metadata param on Axis creation. Defaults to None.

    Raise:
        AssertionError: Raises if any assertion fails.
    """
    assert isinstance(axis, Axis)
    assert value_equal_safe_agains_none_and_nan(axis.transformation_type, input_transformation_type)
    if type(axis.value) is DataField:
        assert_data_field_content(axis.value, input_value, input_unit, input_metadata)
    elif type(axis.value) is TimeSeries:
        assert_timeseries_content(
            axis.value,
            input_value,
            input_unit,
            input_metadata,
            input_timestamps,
            input_timestamps_unit,
            input_timestamps_metadata,
        )
    else:
        raise AssertionError(f"{type(axis.value)} is a invalid type for Axis value property!")


def assert_data_fields_equality(df1: DataField, df2: DataField):
    """Assert two Assonant DataField are equal by comparing their content.

    Args:
        df1 (DataField): DataField which content will be compared to df2.
        df2 (DataField): DataField which content will be compared to df1.
    """
    assert_data_field_content(df1, df2.value, df2.unit, df2.extra_metadata)


def assert_timeseries_equality(ts1: TimeSeries, ts2: TimeSeries):
    """Assert two Assonant TimeSeries are equal by comparing their content.

    Args:
        ts1 (TimeSeries): TimeSeries which content will be compared to ts2.
        ts2 (TimeSeries): TimeSeries which content will be compared to ts1.
    """
    assert_timeseries_content(
        ts1,
        ts2.value.value,
        ts2.value.unit,
        ts2.value.extra_metadata,
        ts2.timestamps.value,
        ts2.value.unit,
        ts2.value.extra_metadata,
    )


def assert_axis_equality(ax1: Axis, ax2: Axis):
    """Assert two Assonant Axis are equal by comparing their content.

    Args:
        ax1 (Axis): Axis which content will be compared to ax2.
        ax2 (Axis): Axis which content will be compared to ax1.
    """

    # If Axis internal DataHandler is not from the same type, the assertion automatically fails
    assert type(ax1.value) is type(ax2.value)

    if type(ax1.value) is DataField:
        assert_axis_content(ax1, ax2.value.value, ax2.value.unit, ax2.value.extra_metadata, ax2.transformation_type)
    elif type(ax1.value) is TimeSeries:
        assert_axis_content(
            ax1,
            ax2.value.value.value,
            ax2.value.value.unit,
            ax2.value.value.extra_metadata,
            ax2.transformation_type,
            ax2.value.timestamps.value,
            ax2.value.timestamps.unit,
            ax2.value.timestamps.extra_metadata,
        )


def assert_component_equality(c1: Component, c2: Component):
    """Assert two Assonant Components are equal by comparing recursively their content dicts.

    Args:
        c1 (Component): Component which content will be compared to c2.
        c2 (Component): Component which content will be compared to c1.
    """

    # If both Components subcomponents keys are not equal, they are clearly different
    assert c1.subcomponents.keys() == c2.subcomponents.keys()

    # Check if Components fields and positions are equal
    assert value_equal_safe_agains_none_and_nan(c1.name, c2.name)
    assert compare_dicts(c1.fields, c2.fields)
    assert compare_dicts(c1.positions, c2.positions)

    # Make a Recursive DFS to verify equality of all subcomponents
    for key in c1.subcomponents.keys():
        assert_component_equality(c1.subcomponents[key], c2.subcomponents[key])


def assert_nxobject_and_assonant_data_field_equality(nxfield: NXfield, data_field: DataField):
    """Assert a NeXus field object and an Assonant DataField are equal by comparing their content.

    Args:
        nxfield (NXfield): NeXus field which content will be compared to the Assonant DataField content.
        data_field (DataField): Assonant DataField which content will be compared to NeXus field content.
    """
    assert array_equal_safe_against_none_and_nan(nxfield.nxvalue, data_field.value)
    for attr_name, attr_value in nxfield.attrs.items():
        if attr_name == "unit":
            assert value_equal_safe_agains_none_and_nan(attr_value.nxvalue, data_field.unit)
        else:
            assert value_equal_safe_agains_none_and_nan(attr_value.nxvalue, data_field.extra_metadata[attr_name])


def assert_nxobject_and_assonant_axis_equality(nxfield: NXfield, axis: Axis):
    """Assert a NeXus field object and an Assonant Axis are equal by comparing their content.

    Args:
        nxfield (NXfield): NeXus field which content will be compared to the Assonant DataField content.
        data_field (Axis): Assonant Axis which content will be compared to NeXus field content.
    """

    if isinstance(axis.value, DataField):
        value_relative_ref = axis.value
    elif isinstance(axis.value, TimeSeries):
        value_relative_ref = axis.value.value
    else:
        raise AssertionError

    assert array_equal_safe_against_none_and_nan(nxfield.nxvalue, value_relative_ref.value)

    for attr_name, attr_value in nxfield.attrs.items():
        if attr_name == "unit":
            assert value_equal_safe_agains_none_and_nan(attr_value.nxvalue, value_relative_ref.unit)
        elif attr_name == "transformation_type":
            assert value_equal_safe_agains_none_and_nan(attr_value.nxvalue, axis.transformation_type.value)
        else:
            assert value_equal_safe_agains_none_and_nan(
                attr_value.nxvalue, value_relative_ref.extra_metadata[attr_name]
            )


def get_list_of_data_handlers_fixture_names(data_handler_type: DataHandler) -> List:
    d = {
        DataField: [
            "example_none_data_field",
            "example_int_1d_array_data_field",
            "example_float_2d_array_data_field",
            "example_string_data_field",
        ],
        Axis: ["example_translation_axis", "example_rotation_axis", "example_none_axis"],
        TimeSeries: [
            "example_int_1d_timeseries_data_field",
            "example_float_2d_timeseries_data_field",
            "example_none_timeseries_data_field",
            "example_timeseries_translation_axis",
            "example_timeseries_rotation_axis",
        ],
    }

    return d[data_handler_type]


def create_test_params_combinations(tested_data_handler_type: DataHandler) -> List:
    """Create params value combination to parametrize DataField creation test.

    Args:
        tested_data_handler_type (DataHandler): Type of DataHandler that will be tested. Based in this,
        the number of parameter returned on each combination tuple vary.

    Raises:
        Exception: Raised when passed tested_data_handler_type is invalid.

    Returns:
        List: List of parameter combination for tests related to DataField creation.
    """

    # Define sets of valid fixture names
    value_fixtures = [
        None,
        "example_int_1d_array",
        "example_float_1d_array",
        "example_float_1d_array_with_nan",
        "example_string_1d_array",
        "example_int_2d_array",
        "example_float_2d_array",
        "example_float_2d_array_with_nan",
        "example_string_2d_array",
        "example_int_1d_list",
        "example_float_1d_list",
        "example_string_1d_list",
        "example_int_1d_list_with_none",
        "example_float_1d_list_with_none",
        "example_string_1d_list_with_none",
        "example_int_2d_list",
        "example_float_2d_list",
        "example_string_2d_list",
        "example_int_2d_list_with_none",
        "example_float_2d_list_with_none",
        "example_string_2d_list_with_none",
    ]

    unit_fixture = [None, "example_unit"]
    metadata_fixture = [None, "example_metadata"]

    if tested_data_handler_type in [DataField, Axis]:
        param_options = [value_fixtures, unit_fixture, metadata_fixture]
    elif tested_data_handler_type == TimeSeries:
        param_options = [
            value_fixtures,
            unit_fixture,
            metadata_fixture,
            value_fixtures,
        ]
    else:
        raise Exception("Passed DataHandler type is invalid")

    # Create all combinations of edge cases based on existing options
    param_combinations = list(itertools.product(*param_options))

    # Combine all
    return param_combinations


def assert_structure_has_keys(data: dict, mandatory: set, optional: set):
    """Assert that a dict has all mandatory keys and only known optional keys."""
    missing = mandatory - data.keys()
    assert not missing, f"Missing mandatory keys: {missing}"

    extra = set(data.keys()) - (mandatory | optional)
    assert not extra, f"Unexpected keys found: {extra}"

    # If a key exists, ensure it has a non-None value
    for key in mandatory:
        assert data[key] is not None, f"Mandatory key '{key}' has None value"
    for key in data.keys() & optional:
        assert data[key] is not None, f"Optional key '{key}' is present but None"


def assert_n_files_exists_in_dir(n_of_files: int, dir_path: Union[Path, str]):
    """
    Assert that a directory contains exactly the expected number of files.

    Args:
        n_of_files (int): The expected number of files in the directory.
        dir_path (Union[Path, str]): Path to the target directory, either as a
            string or a Path object.

    Raises:
        AssertionError: If the number of files found in the directory does not
            match the expected count.
    """
    _dir_path = Path(dir_path) if type(dir_path) is str else dir_path
    files = list(_dir_path.iterdir())
    assert len(files) == n_of_files



"""Exceptions triggered to warn about internal errors related to Assonant metadata retriever classes."""


class DataRetrieverFactoryError(Exception):
    """Data Retriever Factory Error.

    Custom Exception to warn when any error occurs inside DataRetrieverFactory
    """

    pass



"""Implementation of both synchronous and asynchronous message Kafka producer.

See Also:
    https://www.confluent.io/blog/kafka-python-asyncio-integration/
"""
import asyncio
from threading import Thread
from typing import Callable

import confluent_kafka
from confluent_kafka import KafkaException, Message


class Producer:
    """A Multithreading Kafka producer."""

    def __init__(self, configs):
        """Class constructor method.

        Args:
            configs (dict): A dictionary that contains overall kafka cluster configurations.
        """
        self._producer = confluent_kafka.Producer(configs)
        self._cancelled = False
        self._poll_thread = Thread(target=self._poll_loop)
        self._poll_thread.start()

    def _poll_loop(self):
        """Polls the producer for events and calls the corresponding callbacks (if registered)."""
        while not self._cancelled:
            self._producer.poll(0.1)

    def close(self):
        """Wait until all threads terminates.

        This blocks the calling thread until the thread whose join() method is called terminates.
        """
        self._cancelled = True
        self._poll_thread.join()

    def produce(self, topic: str, value: dict, key: str = None, on_delivery: Callable = None):
        """A waitable produce method.

        Initiate sending a message to Kafka, passing in the message value and optionally a key,
        partition, and callback.

        Args:
            topic (str): The name of a topic where the producer will send the message.
            value (dict): Payload of the message.
            key (str): Used to determine the partition within a log to which a message get's appended to.
            on_delivery (Callable, optional): TODO _description_. Defaults to None.
        """
        self._producer.produce(topic, value, on_delivery=on_delivery)


class AsyncProducer:
    """An Awaiting Multithreading Kafka producer."""

    def __init__(self, configs: dict, loop: Callable = None):
        """Class constructor method.

        Args:
            Args:
            configs (dict): A dictionary that contains overall kafka cluster configurations.
            loop (Callable, optional):  A running event loop. Defaults to None.
        """
        self._loop = loop or asyncio.get_event_loop()
        self._producer = confluent_kafka.Producer(configs)
        self._cancelled = False
        self._poll_thread = Thread(target=self._poll_loop)
        self._poll_thread.start()

    def _poll_loop(self):
        """Polls the producer for events and calls the corresponding callbacks (if registered)."""
        while not self._cancelled:
            self._producer.poll(0.1)

    def close(self):
        """Wait until all threads terminates.

        This blocks the calling thread until the thread whose join() method is called terminates.
        """
        self._cancelled = True
        self._poll_thread.join()

    def produce(self, topic: str, value: dict, key: str = None):
        """An awaitable produce method.

        Initiate sending a message to Kafka, passing in the message value and optionally a key,
        partition, and callback.

        Args:
            topic (str): The name of a topic where the producer will send the message.
            value (dict): Payload of the message.
            key (str): Used to determine the partition within a log to which a message get's appended to.

        Returns:
            TODO _type_: TODO _description_
        """
        result = self._loop.create_future()

        def ack(err: str, msg: Message):
            """Delivery report callback to call (from poll() or flush()) on successful or failed delivery.

            A function reference that is called once for each produced message to indicate the final delivery
            result (success or failure).

            Args:
                err (str): Kafka error message.
                msg (Message): The Message object represents either a single consumed or produced message, or an event.
            """
            if err:
                self._loop.call_soon_threadsafe(result.set_exception, KafkaException(err))
            else:
                self._loop.call_soon_threadsafe(result.set_result, msg)

        self._producer.produce(topic, value, on_delivery=ack)
        return result

    def produce_with_delivery(self, topic: str, value: dict, on_delivery: Callable):
        """A produce method with delivery notifications.

        A produce method in which delivery notifications are made available via both the returned
        future and on_delivery callback (if specified).

        Args:
            topic (str): The name of a topic where the producer will send the message.
            value (dict): Payload of the message.
            on_delivery (Callable): Delivery report callback called on successful or failed delivery of message.

        Returns:
            TODO _type_: TODO _description_
        """
        result = self._loop.create_future()

        def ack(err: str, msg: Message):
            """Delivery report callback to call (from poll() or flush()) on successful or failed delivery.

            A function reference that is called once for each produced message to indicate the final delivery
            result (success or failure).

            Args:
                err (str): Kafka error message.
                msg (Message): The Message object represents either a single consumed or produced message, or an event.
            """
            if err:
                self._loop.call_soon_threadsafe(result.set_exception, KafkaException(err))
            else:
                self._loop.call_soon_threadsafe(result.set_result, msg)
            if on_delivery:
                self._loop.call_soon_threadsafe(on_delivery, err, msg)

        self._producer.produce(topic, value, on_delivery=ack)
        return result



"""
Tests focused on validating AssonantDataLogger integration in beamlines

This tests will evaluate AssonantDataLogger functionalities with some real data
and real metadata.csv files from beamlines. The goal here is to verify if nothing will
break when going to production in the beamline.
"""

import json
import os

import pytest

from assonant.data_logger import AssonantDataLogger
from assonant.naming_standards import AcquisitionType, BeamlineName, ExperimentStage


@pytest.mark.parametrize("beamline_name", [beamline_name for beamline_name in BeamlineName])
@pytest.mark.forked  # This is MANDATORY for this test to correctly isolate env variables for EPICS_CA_ADDR_LIST
def test_log_nexus_file_from_real_bluesky_data_files_from_files(
    beamline_name, real_csv_metadata_file_path_map, tmp_path, real_bluesky_data_test_cases_dir_path_map
):
    """
    Validate that AssonantDataLogger works correctly when using real beamline
    csv metadata to log real beamline bluesky data if directly accessing
    bluesky json files.

    PS: Content is not being validated only if module does not break!

    Expected Behavior:
        - Tests must be manually evaluated as expected behavior may vary among test cases, but
        it is important to check if overall generate structure within nexus file is standardized
        as expected.
    """
    if not os.path.exists(real_csv_metadata_file_path_map[beamline_name.value]):
        pytest.skip(reason=f"CSV metadata for {beamline_name} does not exist yet!")
    elif beamline_name.value not in real_bluesky_data_test_cases_dir_path_map.keys():
        pytest.skip(reason=f"There is no test case with real data for {beamline_name} beamline yet!")
    else:
        log_file_name = f"test_{beamline_name.value}_file"

        data_logger = AssonantDataLogger(
            beamline_name,
            data_schema_file_path=real_csv_metadata_file_path_map[beamline_name.value],
        )

        for i in range(0, len(real_bluesky_data_test_cases_dir_path_map[beamline_name.value])):
            # Execute test for each test case dir that exist for that beamline
            test_case_dir = real_bluesky_data_test_cases_dir_path_map[beamline_name.value][i]
            log_file_path = tmp_path / f"test_case_{i}"

            for test_case_file in os.scandir(test_case_dir):
                if test_case_file.is_file() and test_case_file.name.endswith(".json"):
                    data_logger.log_bluesky_data_from_json_file(
                        log_file_path,
                        log_file_name,
                        ExperimentStage.SAMPLE_ACQUISITION,
                        AcquisitionType.SNAPSHOT_AT_START,
                        test_case_file.path,
                    )
                    data_logger.log_bluesky_data_from_json_file(
                        log_file_path,
                        log_file_name,
                        ExperimentStage.SAMPLE_ACQUISITION,
                        AcquisitionType.SCAN,
                        test_case_file.path,
                    )
                    data_logger.log_bluesky_data_from_json_file(
                        log_file_path,
                        log_file_name,
                        ExperimentStage.SAMPLE_ACQUISITION,
                        AcquisitionType.SNAPSHOT_AT_END,
                        test_case_file.path,
                    )


@pytest.mark.parametrize("beamline_name", [beamline_name for beamline_name in BeamlineName])
@pytest.mark.forked  # This is MANDATORY for this test to correctly isolate env variables for EPICS_CA_ADDR_LIST
def test_log_nexus_file_from_real_bluesky_data_files_from_documents_list(
    beamline_name, real_csv_metadata_file_path_map, tmp_path, real_bluesky_data_test_cases_dir_path_map
):
    """
    Validate that AssonantDataLogger works correctly when using real beamline
    csv metadata to log real beamline bluesky data if directly accessing
    bluesky json files.

    PS: Content is not being validated only if module does not break!

    Expected Behavior:
        - Tests must be manually evaluated as expected behavior may vary among test cases, but
        it is important to check if overall generate structure within nexus file is standardized
        as expected.
    """
    if not os.path.exists(real_csv_metadata_file_path_map[beamline_name.value]):
        pytest.skip(reason=f"CSV metadata for {beamline_name} does not exist yet!")
    elif beamline_name.value not in real_bluesky_data_test_cases_dir_path_map.keys():
        pytest.skip(reason=f"There is no test case with real data for {beamline_name} beamline yet!")
    else:
        log_file_name = f"test_{beamline_name.value}_file"

        data_logger = AssonantDataLogger(
            beamline_name,
            data_schema_file_path=real_csv_metadata_file_path_map[beamline_name.value],
        )

        for i in range(0, len(real_bluesky_data_test_cases_dir_path_map[beamline_name.value])):
            # Execute test for each test case dir that exist for that beamline
            test_case_dir = real_bluesky_data_test_cases_dir_path_map[beamline_name.value][i]
            log_file_path = tmp_path / f"test_case_{i}"

            for test_case_file in os.scandir(test_case_dir):
                if test_case_file.is_file() and test_case_file.name.endswith(".json"):

                    with open(test_case_file) as f:
                        test_case_file_data = json.load(f)

                    data_logger.log_bluesky_data_from_documents_list(
                        log_file_path,
                        log_file_name,
                        ExperimentStage.SAMPLE_ACQUISITION,
                        AcquisitionType.SNAPSHOT_AT_START,
                        test_case_file_data,
                    )
                    data_logger.log_bluesky_data_from_documents_list(
                        log_file_path,
                        log_file_name,
                        ExperimentStage.SAMPLE_ACQUISITION,
                        AcquisitionType.SCAN,
                        test_case_file_data,
                    )
                    data_logger.log_bluesky_data_from_documents_list(
                        log_file_path,
                        log_file_name,
                        ExperimentStage.SAMPLE_ACQUISITION,
                        AcquisitionType.SNAPSHOT_AT_END,
                        test_case_file_data,
                    )



"""Assonant data classes - Components.

Data classes that defines available Assonant components.
"""

from .aperture import Aperture
from .attenuator import Attenuator
from .beam import Beam
from .beam_stopper import BeamStopper
from .beamline import Beamline
from .bending_magnet import BendingMagnet
from .bvs import BVS
from .collimator import Collimator
from .component import Component
from .cryojet import Cryojet
from .detector import Detector, DetectorChannel, DetectorModule, DetectorROI
from .dewar import Dewar
from .entry import Entry
from .fresnel_zone_plate import FresnelZonePlate
from .granite_base import GraniteBase
from .grating import Grating
from .mirror import Mirror
from .monochromator import (
    Monochromator,
    MonochromatorCrystal,
    MonochromatorVelocitySelector,
)
from .pinhole import Pinhole
from .sample import Sample
from .sensor import Sensor
from .shutter import Shutter
from .slit import Slit
from .storage_ring import StorageRing
from .undulator import Undulator
from .wiggler import Wiggler

__all__ = [
    "Aperture",
    "Attenuator",
    "Beam",
    "BeamStopper",
    "Beamline",
    "BendingMagnet",
    "BVS",
    "Collimator",
    "Component",
    "Cryojet",
    "Detector",
    "DetectorModule",
    "DetectorROI",
    "DetectorChannel",
    "Dewar",
    "Entry",
    "FresnelZonePlate",
    "GraniteBase",
    "Grating",
    "Mirror",
    "Monochromator",
    "MonochromatorCrystal",
    "MonochromatorVelocitySelector",
    "Pinhole",
    "Sample",
    "Sensor",
    "Shutter",
    "Slit",
    "StorageRing",
    "Undulator",
    "Wiggler",
]



"""Assonant metadata retriever.

This submodule defines Metadata Retrievers objects to allow retrieving important
metadata for Assonant from different sources.
"""

from .assonant_metadata_retriever import AssonantMetadataRetriever
from .assonant_metadata_retriever_interface import IAssonantMetadataRetriever
from .exceptions import AssonantMetadataRetrieverError
from .metadata_retriever_factory import MetadataRetrieverFactoryError

__all__ = [
    "AssonantMetadataRetriever",
    "AssonantMetadataRetrieverError",
    "IAssonantMetadataRetriever",
    "MetadataRetrieverFactoryError",
]



"""Exceptions triggered to warn about internal errors in File Writer Factory module."""


class FileWriterFactoryError(Exception):
    """File Writer Factory Error.

    Custom Exception to warn when any error occurs inside FileWriterFactory
    """

    pass



"""Exceptions triggered to warn about internal errors in Assonant File Writer module."""


class AssonantFileWriterError(Exception):
    """Assonant File Writer Error.

    Custom Exception to warn when any error occurs inside AssonantFileWriter
    """

    pass


class FileAlreadyExistsError(Exception):
    """File Writer Error.

    Custom Exception to warn when the file already exists when trying to write new data.
    """

    pass


class NeXusObjectAlreadyExistsError(Exception):
    """NeXus File Writer Error.

    Custom Exception to warn when trying to write an NeXus object that already exist in the NeXus file.

    """



from enum import Enum


class AcquisitionType(Enum):
    """Possible Acquisition Types values."""

    SNAPSHOT_AT_START = "snapshot_at_start"
    SNAPSHOT_AT_END = "snapshot_at_end"
    SCAN = "scan"



import json
import os

import pytest

from assonant._bluesky.data_parser import DocumentsDataParser
from assonant.naming_standards import BeamlineName


def assert_query_config_by_pv_name_structure(config_dict: dict) -> None:
    """
    Validate structure returned by query_config_by_pv_name() from bluesky data parser.
    """
    assert isinstance(config_dict, dict), "Configs must be a dict"

    for pv_name, pv_info in config_dict.items():
        assert isinstance(pv_name, str), "Config PV name must be str"
        assert isinstance(pv_info, dict), f"{pv_name} value must be dict"

        required_keys = {"value", "timestamps"}
        assert required_keys == set(pv_info.keys()), f"{pv_name} must contain keys {required_keys}"


def assert_query_data_by_stream_structure(data_dict: dict, config_may_exist: bool = True) -> None:
    """
    Validate structure returned by query_data_by_stream_name() from bluesky data parser.
    """
    assert isinstance(data_dict, dict), "Stream data must be a dict"

    for seq_num, seq_dict in data_dict.items():
        assert isinstance(seq_num, int), f"{seq_num} must be an int"
        assert isinstance(seq_dict, dict), f"{seq_num} must map to dict"

        for pv_name, pv_info in seq_dict.items():
            assert isinstance(pv_name, str), f"{pv_name} must be an str"
            assert isinstance(pv_info, dict), f"{pv_name} must map to dict"

            required_keys = {"data"}
            assert required_keys.issubset(pv_info.keys()), f"{pv_name} must contain keys {required_keys}"

            # validate data block
            data_block = pv_info["data"]
            assert isinstance(data_block, dict), f"{pv_name} data must be dict"

            data_required = {"value", "timestamps"}
            assert data_required == set(data_block.keys()), f"{pv_name} data must contain {data_required}"

            # validate configs block
            if config_may_exist is True:
                if "configs" in pv_info.keys():
                    assert_query_config_by_pv_name_structure(pv_info["configs"])
                # else: Nothing happens as "configs" is optional
            else:
                assert "configs" not in pv_info.keys(), f"{pv_name} must not contain configs block in this test case"


def assert_query_all_data_structure(all_data: dict, config_may_exist: bool = True) -> None:
    """
    Validate structure returned by query_all_data() from bluesky data parser.
    """
    assert isinstance(all_data, dict), "All data must be dict"

    for stream_name, stream_data in all_data.items():
        assert isinstance(stream_name, str), "Stream name must be str"
        assert isinstance(stream_data, dict), "Stream data must map to dict"

        assert_query_data_by_stream_structure(stream_data, config_may_exist)


@pytest.fixture(scope="session")
def example_bluesky_data_parser(example_bluesky_all_streams_acquisition_file_path) -> DocumentsDataParser:
    """
    Provides a functional DocumentsDataParser for the example_bluesky_all_streams_acquisition_file.json
    """

    with open(example_bluesky_all_streams_acquisition_file_path) as f:
        example_bluesky_docs_list = json.load(f)

    return DocumentsDataParser(bluesky_documents_list=example_bluesky_docs_list)


def test_query_all_data(example_bluesky_data_parser):
    """
    Validate that query_all_data method of DocumentsDataParser works correctly and don't break when using
    it to retrieve data from the example_bluesky_all_streams_acquisition_file.json file.

    PS: Only types and structure is being checked, not values!

    Expected Behavior:
        - The returned value must be a dictionary with the structure described in the method documentation.
        - The returned value must contain data related to all Acquisition Types (Primary, Baseline and Monitor)
    """
    retrieved_data = example_bluesky_data_parser.query_all_data(with_config=True)
    # Assert that the returned structure is correct
    assert_query_all_data_structure(retrieved_data, config_may_exist=True)


def test_query_data_by_stream_name(example_bluesky_data_parser):
    """
    Validate that query_data_by_stream_name method of DocumentsDataParser works correctly and don't break when using
    it to retrieve data from the example_bluesky_all_streams_acquisition_file.json file.

    PS: Only types and structure is being checked, not values!

    Expected Behavior:
        - The returned value must be a dictionary with the structure described in the method documentation.
        - The returned value must contain data related to the specified Acquisition Type
    """
    for stream_name in example_bluesky_data_parser.get_stream_names():

        retrieved_data = example_bluesky_data_parser.query_data_by_stream_name(event_stream_name=stream_name)

        # Assert that the returned structure is correct
        assert_query_data_by_stream_structure(retrieved_data, config_may_exist=True)


def test_query_config_by_pv_name(example_bluesky_data_parser):
    """
    Validate that query_config_by_pv_name method of DocumentsDataParser works correctly and don't break when using
    it to retrieve data from the example_bluesky_all_streams_acquisition_file.json file.

    PS: Only types and structure is being checked, not values!

    Expected Behavior:
        - The returned value must be a dictionary with the structure described in the
        method documentation. (Check if assertion method is correctly reflecting it!)
        - The returned value must contain config information for all PVs related to all Acquisition Types
    """
    for stream_name in example_bluesky_data_parser.get_stream_names():

        stream_data = example_bluesky_data_parser.query_data_by_stream_name(event_stream_name=stream_name)

        for pv_name in stream_data.keys():
            retrieved_pv_config = example_bluesky_data_parser.query_config_by_pv_name(
                pv_name=pv_name, stream_name=stream_name
            )
            # Assert that the returned structure is correct
            assert_query_config_by_pv_name_structure(retrieved_pv_config)


def test_query_all_data_without_config(example_bluesky_data_parser):
    """
    Validate that query_all_data method of DocumentsDataParser works correctly and don't break when using
    it to retrieve data from the example_bluesky_all_streams_acquisition_file.json without config metadata.

    PS: Only types and structure is being checked, not values!

    Expected Behavior:
        - Returned structure must be the same as query_all_data with config but without the "configs" block in
    the innermost dicts.
    """
    retrieved_data = example_bluesky_data_parser.query_all_data(with_config=False)
    # Assert that the returned structure is correct
    assert_query_all_data_structure(retrieved_data, config_may_exist=False)


def test_query_data_by_stream_name_without_config(example_bluesky_data_parser):
    """
    Validate that query_data_by_stream_name method of DocumentsDataParser works correctly and don't break when using
    it to retrieve data from the example_bluesky_all_streams_acquisition_file.json without config metadata.

    PS: Only types and structure is being checked, not values!

    Expected Behavior:
        - Returned structure must be the same as query_data_by_stream_name with config but without the
        "configs" block in the innermost dicts.
    """
    for stream_name in example_bluesky_data_parser.get_stream_names():

        retrieved_data = example_bluesky_data_parser.query_data_by_stream_name(
            event_stream_name=stream_name, with_config=False
        )

        # Assert that the returned structure is correct
        assert_query_data_by_stream_structure(retrieved_data, config_may_exist=False)


@pytest.mark.parametrize("beamline_name", [beamline_name for beamline_name in BeamlineName])
@pytest.mark.forked
def test_parse_real_data_with_config_from_beamline_example_bluesky_files(
    beamline_name, tmp_path, real_bluesky_data_test_cases_dir_path_map
):
    """
    Validate that DataParser works correctly and don't break when using
    it to retrieve data from beamline real bluesky example JSON files.

    PS: Only types and structure is being checked, not values!

    Expected Behavior:
        - Returned structure for each case must be the same no matter from which beamline it comes.
        - Returned structure must be the same as docummented on method docstring for
        query_all_data() method. (Check if assertion method is up-to-date!)
        - Returned structure MUST contain config blocks which also must be checked by the
        assertion method.
    """
    if beamline_name.value not in real_bluesky_data_test_cases_dir_path_map.keys():
        pytest.skip(reason=f"There is no test case with real data for {beamline_name} beamline yet!")
    else:
        for i in range(0, len(real_bluesky_data_test_cases_dir_path_map[beamline_name.value])):
            # Execute test for each test case dir that exist for that beamline
            test_case_dir = real_bluesky_data_test_cases_dir_path_map[beamline_name.value][i]

            for test_case_file in os.scandir(test_case_dir):
                if test_case_file.is_file() and test_case_file.name.endswith(".json"):

                    with open(test_case_file) as f:
                        test_case_bluesky_docs_list = json.load(f)

                    doc_parser = DocumentsDataParser(bluesky_documents_list=test_case_bluesky_docs_list)

                    # Assert that Data Parser can retrieve all data with config without error
                    all_data = doc_parser.query_all_data(with_config=True)
                    assert_query_all_data_structure(all_data, config_may_exist=True)


@pytest.mark.parametrize("beamline_name", [beamline_name for beamline_name in BeamlineName])
@pytest.mark.forked
def test_parse_real_data_without_config_from_beamline_example_bluesky_files(
    beamline_name, tmp_path, real_bluesky_data_test_cases_dir_path_map
):
    """
    Validate that DataParser works correctly and don't break when using it to retrieve
    data from beamline real bluesky example JSON files without config metadata.

    PS: Only types and structure is being checked, not values!

    Expected Behavior:
        - Returned structure for each case must be the same no matter from which beamline it comes.
        - Returned structure must be the same as docummented on method docstring for
        query_all_data() method. (Check if assertion method is up-to-date!)
        - Returned structure must NOT contain config blocks which also must be checked by the
        assertion method.
    """
    if beamline_name.value not in real_bluesky_data_test_cases_dir_path_map.keys():
        pytest.skip(reason=f"There is no test case with real data for {beamline_name} beamline yet!")
    else:

        for i in range(0, len(real_bluesky_data_test_cases_dir_path_map[beamline_name.value])):
            # Execute test for each test case dir that exist for that beamline
            test_case_dir = real_bluesky_data_test_cases_dir_path_map[beamline_name.value][i]

            for test_case_file in os.scandir(test_case_dir):
                if test_case_file.is_file() and test_case_file.name.endswith(".json"):

                    with open(test_case_file) as f:
                        test_case_bluesky_docs_list = json.load(f)

                    doc_parser = DocumentsDataParser(bluesky_documents_list=test_case_bluesky_docs_list)

                    # Assert that Data Parser can retrieve all data with config without error
                    all_data = doc_parser.query_all_data(with_config=False)
                    assert_query_all_data_structure(all_data, config_may_exist=False)



import importlib
import inspect
import random
from collections import defaultdict
from pathlib import Path
from typing import Any, Dict, Generator, List, Union

import numpy as np
import pandas as pd
import pytest
from nexusformat.nexus import *

from assonant.data_classes.components import *
from assonant.data_classes.components import Component
from assonant.data_classes.data_handlers import Axis, DataField, DataHandler, TimeSeries
from assonant.data_classes.enums import TransformationType
from assonant.data_classes.factories import (
    AssonantComponentFactory,
    AssonantDataHandlerFactory,
)
from assonant.naming_standards import BeamlineName


@pytest.fixture(scope="session")
def real_bluesky_data_test_cases_dir_path_map():
    """
    Provides a pytest fixture returning a mapping where the key is the beamline
    name (str) and the value is a list of directories containing the real test
    data for that beamline.

    Example:
        {
            "carnauba": [Path(".../carnauba_test_3")],
            "manaca": [Path(".../manaca_test_4")],
            "sapucaia": [
                Path(".../sapucaia_test_1"),
                Path(".../sapucaia_test_2")
            ],
        }
    """
    base_dir = Path(__file__).parent / "data" / "real_data" / "bluesky"
    path_mapping = defaultdict(list)

    for subdir in base_dir.iterdir():
        if subdir.is_dir() and "_test_" in subdir.name:
            beamline = subdir.name.split("_test_")[0]
            path_mapping[beamline.upper()].append(subdir)

    return dict(path_mapping)


@pytest.fixture(scope="session")
def real_csv_metadata_file_path_map() -> str:
    """
    Provides as a pytest fixture a dict where the key is the beamline name and
    the value is the respective filepath for the real metadata.csv files for [
    integration tests.
    """
    path_mapping = {}
    for beamline_name in BeamlineName:
        path_mapping[beamline_name.value] = str(
            Path(__file__).parent / "data" / "real_data" / "csv" / beamline_name.value.lower() / "metadata.csv"
        )
    return path_mapping


@pytest.fixture(scope="session")
def example_csv_metadata_file_path() -> str:
    """
    Provides as a pytest fixture the path for the example_metadata.csv file used for tests.
    """
    return str(Path(__file__).parent / "data" / "dummy_data" / "example_metadata.csv")


@pytest.fixture(scope="session")
def example_bluesky_baseline_acquisition_file_path() -> str:
    """
    Provides as a pytest fixture the path for the example_bluesky_baseline_acquisition.json
    file used for tests, that data simulating a bluesky baseline acquisition.
    """
    return str(Path(__file__).parent / "data" / "dummy_data" / "example_bluesky_baseline_acquisition.json")


@pytest.fixture(scope="session")
def example_bluesky_primary_acquisition_file_path() -> str:
    """
    Provides as a pytest fixture the path for the example_bluesky_primary_acquisition.json file used for tests.
    """
    return str(Path(__file__).parent / "data" / "dummy_data" / "example_bluesky_primary_acquisition.json")


@pytest.fixture(scope="session")
def example_bluesky_monitor_acquisition_file_path() -> str:
    """
    Provides as a pytest fixture the path for the example_bluesky_monitor_acquisition.json file used for tests.
    """
    return str(Path(__file__).parent / "data" / "dummy_data" / "example_bluesky_monitor_acquisition.json")


@pytest.fixture(scope="session")
def example_bluesky_flyscan_acquisition_file_path() -> str:
    """
    Provides as a pytest fixture the path for the example_bluesky_monitor_acquisition.json file used for tests.
    """
    return str(Path(__file__).parent / "data" / "dummy_data" / "example_bluesky_flyscan_acquisition.json")


@pytest.fixture(scope="session")
def example_bluesky_all_streams_acquisition_file_path() -> str:
    """
    Provides as a pytest fixture the path for the example_bluesky_all_streams_acquisition.json
    file used for tests, that data simulating a bluesky acquisition with data passed through
    baseline, primary and monitor streams.
    """
    return str(Path(__file__).parent / "data" / "dummy_data" / "example_bluesky_all_streams_acquisition.json")


@pytest.fixture(scope="session")
def example_metadata_csv_df(example_csv_metadata_file_path):
    return pd.read_csv(example_csv_metadata_file_path, skiprows=1)


@pytest.fixture(scope="session")
def fixed_seed_random_generator() -> Generator:
    """
    Provides a NumPy random number generator with a fixed seed (42) to ensure reproducibility across tests.
    """
    return np.random.default_rng(seed=42)


# ======= Non-array data =======


@pytest.fixture(scope="session")
def example_int_value() -> int:
    """
    Provides a sample integer value for basic test cases.
    """
    return 1234


@pytest.fixture(scope="session")
def example_float_value() -> float:
    """
    Provides a sample float value for basic test cases.
    """
    return 43.21


@pytest.fixture(scope="session")
def example_string_value() -> float:
    """
    Provides a sample string value for basic test cases.
    """
    return "example value"


# ======= 1D data as numpy arrays =======


@pytest.fixture(scope="session")
def example_int_1d_array(fixed_seed_random_generator) -> np.typing.NDArray:
    """
    Provides a 1D NumPy array of integers (length 10) with values from 0 to 255.
    """
    return fixed_seed_random_generator.integers(0, 256, 10)


@pytest.fixture(scope="session")
def example_float_1d_array(fixed_seed_random_generator) -> np.typing.NDArray:
    """
    Provides a 1D NumPy array of floats (length 10) with values in [0, 1).
    """
    return fixed_seed_random_generator.random(10)


@pytest.fixture(scope="session")
def example_string_1d_array() -> np.typing.NDArray:
    """
    Provides a 1D NumPy array of sample string values.
    """
    return np.array(["a", "b", "c", "d", "e"])


@pytest.fixture(scope="session")
def example_float_1d_array_with_nan(example_float_1d_array) -> np.typing.NDArray:
    """
    Provides a copy of a float 1D NumPy array with NaNs injected at the start, middle, and end.
    Useful for testing NaN-safe operations.
    """
    float_1d_array_with_nan = example_float_1d_array.copy()
    float_1d_array_with_nan[0] = np.nan
    float_1d_array_with_nan[len(float_1d_array_with_nan) // 2] = np.nan
    float_1d_array_with_nan[-1] = np.nan
    return float_1d_array_with_nan


# ======= 2D data as numpy arrays =======


@pytest.fixture(scope="session")
def example_int_2d_array(fixed_seed_random_generator) -> np.typing.NDArray:
    """
    Provides a 10x10 NumPy array of random integers in [0, 256).
    """
    return fixed_seed_random_generator.integers(0, 256, (10, 10))


@pytest.fixture(scope="session")
def example_float_2d_array(fixed_seed_random_generator) -> np.typing.NDArray:
    """
    Provides a 10x10 NumPy array of random floats in [0, 1).
    """
    return fixed_seed_random_generator.random((10, 10))


@pytest.fixture(scope="session")
def example_string_2d_array() -> np.typing.NDArray:
    """
    Provides a 2D NumPy array (4x5) of sample strings.
    """
    return np.array(
        [["a", "b", "c", "d", "e"], ["f", "g", "h", "i", "j"], ["k", "l", "m", "n", "o"], ["p", "q", "r", "s", "t"]]
    )


@pytest.fixture(scope="session")
def example_float_2d_array_with_nan(example_float_2d_array) -> np.typing.NDArray:
    """
    Provides a copy of a float 2D NumPy array with NaNs injected at the top-left, center, and bottom-right.
    Useful for testing NaN handling in 2D data.
    """
    float_2d_array_with_nan = example_float_2d_array.copy()
    float_2d_array_with_nan[(0, 0)] = np.nan
    float_2d_array_with_nan[(len(float_2d_array_with_nan) // 2, len(float_2d_array_with_nan) // 2)] = np.nan
    float_2d_array_with_nan[(-1, -1)] = np.nan
    return float_2d_array_with_nan


# ======= 1D data as python list =======


@pytest.fixture(scope="session")
def example_int_1d_list() -> List[int]:
    """
    Provides a 1D list of 10 random integers in [0, 255] with fixed seed.
    """
    random.seed(42)
    return [random.randint(0, 255) for _ in range(10)]


@pytest.fixture(scope="session")
def example_float_1d_list() -> List[float]:
    """
    Provides a 1D list of 10 random float values in [0, 1) with fixed seed.
    """
    random.seed(42)
    return [random.random() for _ in range(10)]


@pytest.fixture(scope="session")
def example_string_1d_list() -> List[str]:
    """
    Provides a 1D list of sample string values.
    """
    return ["a", "b", "c", "d", "e"]


@pytest.fixture(scope="session")
def example_int_1d_list_with_none(example_int_1d_list) -> List[Union[int, None]]:
    """
    Provides a copy of the int list with None values at the start, middle, and end.
    """
    int_1d_list_with_none = example_int_1d_list.copy()
    int_1d_list_with_none[0] = None
    int_1d_list_with_none[len(int_1d_list_with_none) // 2] = None
    int_1d_list_with_none[-1] = None
    return int_1d_list_with_none


@pytest.fixture(scope="session")
def example_float_1d_list_with_none(example_float_1d_list) -> List[Union[float, None]]:
    """
    Provides a copy of the float list with None values at the start, middle, and end.
    """
    float_1d_list_with_none = example_float_1d_list.copy()
    float_1d_list_with_none[0] = None
    float_1d_list_with_none[len(float_1d_list_with_none) // 2] = None
    float_1d_list_with_none[-1] = None
    return float_1d_list_with_none


@pytest.fixture(scope="session")
def example_string_1d_list_with_none(example_string_1d_list) -> List[Union[str, None]]:
    """
    Provides a copy of the string list with None values at the start, middle, and end.
    """
    string_1d_list_with_none = example_string_1d_list.copy()
    string_1d_list_with_none[0] = None
    string_1d_list_with_none[len(string_1d_list_with_none) // 2] = None
    string_1d_list_with_none[-1] = None
    return string_1d_list_with_none


# ======= 2D data as python list =======


@pytest.fixture(scope="session")
def example_int_2d_list() -> List[List[int]]:
    """
    Provides a 2D list (10x10) of random integers in [0, 255] with fixed seed.
    """
    random.seed(42)
    return [[random.randint(0, 255) for _ in range(10)] for _ in range(10)]


@pytest.fixture(scope="session")
def example_float_2d_list() -> List[List[float]]:
    """
    Provides a 2D list (10x10) of random float values in [0, 1) with fixed seed.
    """
    random.seed(42)
    return [[random.random() for _ in range(10)] for _ in range(10)]


@pytest.fixture(scope="session")
def example_string_2d_list() -> List[List[str]]:
    """
    Provides a 2D list of sample string values (4x5 grid).
    """
    return [["a", "b", "c", "d", "e"], ["f", "g", "h", "i", "j"], ["k", "l", "m", "n", "o"], ["p", "q", "r", "s", "t"]]


@pytest.fixture(scope="session")
def example_int_2d_list_with_none(example_int_2d_list) -> List[List[Union[int, None]]]:
    """
    Provides a copy of the 2D int list with None values injected at the top-left, center, and bottom-right.
    """
    int_2d_list_with_none = example_int_2d_list.copy()
    int_2d_list_with_none[0][0] = None
    int_2d_list_with_none[len(int_2d_list_with_none) // 2][len(int_2d_list_with_none) // 2] = None
    int_2d_list_with_none[-1][-1] = None
    return int_2d_list_with_none


@pytest.fixture(scope="session")
def example_float_2d_list_with_none(example_float_2d_list) -> List[List[Union[float, None]]]:
    """
    Provides a copy of the 2D float list with None values injected at the top-left, center, and bottom-right.
    """
    float_2d_list_with_none = example_float_2d_list.copy()
    float_2d_list_with_none[0][0] = None
    float_2d_list_with_none[len(example_float_2d_list) // 2][len(example_float_2d_list) // 2] = None
    float_2d_list_with_none[-1][-1] = None
    return float_2d_list_with_none


@pytest.fixture(scope="session")
def example_string_2d_list_with_none(example_string_2d_list) -> List[List[Union[str, None]]]:
    """
    Provides a copy of the 2D string list with None values injected at the top-left, center, and bottom-right.
    """
    string_2d_list_with_none = example_string_2d_list.copy()
    string_2d_list_with_none[0][0] = None
    string_2d_list_with_none[len(string_2d_list_with_none) // 2][len(string_2d_list_with_none) // 2] = None
    string_2d_list_with_none[-1][-1] = None
    return string_2d_list_with_none


# ======= Additional metadata =======


@pytest.fixture(scope="session")
def example_metadata(fixed_seed_random_generator) -> Dict[str, Any]:
    """
    Provides a sample metadata dictionary with mixed types, including NumPy arrays.
    Useful for testing metadata compatibility and serialization.
    """
    return {
        "str metadata": "I'm a string metadata",
        "float metadata": 1.5,
        "int metadata": 42,
        "list metadata": [1, 2, 3, 4, 5],
        "numpy array metadata": fixed_seed_random_generator.integers(0, 10, 5),
    }


@pytest.fixture(scope="session")
def example_unit() -> str:
    """
    Provides a sample unit string representing a measurement unit.
    """
    return "Some measurement unit"


# ======= Assonant Entries =======


@pytest.fixture(scope="function")
def example_entry() -> Entry:
    """
    Provides a mocked Assonant entry with a fixed name.
    """
    return AssonantComponentFactory.create_component_by_class_type(Entry, component_name="entry")


@pytest.fixture(scope="function")
def example_beamline() -> Beamline:
    """
    Provides a mocked Assonant entry with a fixed name.
    """
    return AssonantComponentFactory.create_component_by_class_type(Beamline, component_name="beamline")


# ======= Assonant DataHandlers =======


# DataField
@pytest.fixture(scope="function")
def example_none_data_field() -> DataField:
    """
    Provides a mocked data field with all values set to None.
    """
    return AssonantDataHandlerFactory.create_data_field(value=None, unit=None, extra_metadata=None)


@pytest.fixture(scope="function")
def example_int_1d_array_data_field(example_int_1d_array, example_unit, example_metadata) -> DataField:
    """
    Provides a mocked data field using a 1D integer array and metadata.
    """
    return AssonantDataHandlerFactory.create_data_field(
        value=example_int_1d_array, unit=example_unit, extra_metadata=example_metadata
    )


@pytest.fixture(scope="function")
def example_float_2d_array_data_field(example_float_2d_array, example_unit, example_metadata) -> DataField:
    """
    Provides a mocked data field using a 2D float array and metadata.
    """
    return AssonantDataHandlerFactory.create_data_field(
        value=example_float_2d_array, unit=example_unit, extra_metadata=example_metadata
    )


@pytest.fixture(scope="function")
def example_string_data_field(example_string_1d_array, example_unit, example_metadata) -> DataField:
    """
    Provides a mocked data field using a 1D string array and metadata.
    """
    return AssonantDataHandlerFactory.create_data_field(
        value=example_string_1d_array, unit=example_unit, extra_metadata=example_metadata
    )


# Axis
@pytest.fixture(scope="function")
def example_translation_axis(example_float_1d_array, example_unit, example_metadata) -> Axis:
    """
    Provides a mocked translation axis using a 1D float array and metadata.
    """
    return AssonantDataHandlerFactory.create_axis(
        transformation_type=TransformationType.TRANSLATION,
        value=example_float_1d_array,
        unit=example_unit,
        extra_metadata=example_metadata,
    )


@pytest.fixture(scope="function")
def example_rotation_axis(example_float_1d_array, example_unit, example_metadata) -> Axis:
    """
    Provides a mocked rotation axis using a 1D float array and metadata.
    """
    return AssonantDataHandlerFactory.create_axis(
        transformation_type=TransformationType.ROTATION,
        value=example_float_1d_array,
        unit=example_unit,
        extra_metadata=example_metadata,
    )


@pytest.fixture(scope="function")
def example_none_axis() -> Axis:
    """
    Provides a mocked rotation axis using a 1D float array and metadata.
    """
    return AssonantDataHandlerFactory.create_axis(
        transformation_type=TransformationType.ROTATION,
        value=None,
        unit=None,
        extra_metadata=None,
    )


# TimeSeries
@pytest.fixture(scope="function")
def example_int_1d_timeseries_data_field(
    example_float_1d_array, example_int_1d_array, example_unit, example_metadata
) -> TimeSeries:
    """
    Provides a mocked timeseries data field using a 1D float array and integer timestamps.
    """
    return AssonantDataHandlerFactory.create_timeseries_field(
        value=example_float_1d_array,
        timestamps=example_int_1d_array,
        unit=example_unit,
        extra_metadata=example_metadata,
        timestamps_unit=example_unit,
        timestamp_extra_metadata=example_metadata,
    )


@pytest.fixture(scope="function")
def example_float_2d_timeseries_data_field(
    example_float_2d_array, example_int_1d_array, example_unit, example_metadata
) -> TimeSeries:
    """
    Provides a mocked timeseries data field using a 2D float array and integer timestamps.
    """
    return AssonantDataHandlerFactory.create_timeseries_field(
        value=example_float_2d_array,
        timestamps=example_int_1d_array,
        unit=example_unit,
        extra_metadata=example_metadata,
        timestamps_unit=example_unit,
        timestamp_extra_metadata=example_metadata,
    )


@pytest.fixture(scope="function")
def example_none_timeseries_data_field() -> TimeSeries:
    """
    Provides a mocked timeseries data field using a 1D float array and integer timestamps.
    """
    return AssonantDataHandlerFactory.create_timeseries_field(
        value=None,
        timestamps=None,
        unit=None,
        extra_metadata=None,
        timestamps_unit=None,
        timestamp_extra_metadata=None,
    )


@pytest.fixture(scope="function")
def example_timeseries_translation_axis(
    example_float_1d_array, example_int_1d_array, example_unit, example_metadata
) -> Axis:
    """
    Provides a mocked timeseries translation axis using 1D float values and integer timestamps.
    """
    return AssonantDataHandlerFactory.create_timeseries_axis(
        transformation_type=TransformationType.TRANSLATION,
        value=example_float_1d_array,
        timestamps=example_int_1d_array,
        unit=example_unit,
        extra_metadata=example_metadata,
        timestamps_unit=example_unit,
        timestamp_extra_metadata=example_metadata,
    )


@pytest.fixture(scope="function")
def example_timeseries_rotation_axis(
    example_float_2d_array, example_int_1d_array, example_unit, example_metadata
) -> Axis:
    """
    Provides a mocked timeseries rotation axis using 2D float values and integer timestamps.
    """
    return AssonantDataHandlerFactory.create_timeseries_axis(
        transformation_type=TransformationType.ROTATION,
        value=example_float_2d_array,
        timestamps=example_int_1d_array,
        unit=example_unit,
        extra_metadata=example_metadata,
        timestamps_unit=example_unit,
        timestamp_extra_metadata=example_metadata,
    )


# ======= Assonant Components and Subcomponents =======


@pytest.fixture(scope="function")
def example_base_component() -> Component:
    """
    Provides a mocked base component named 'main_component' using the generic Component class.
    """
    return AssonantComponentFactory.create_component_by_class_type(Component, "main_component")


@pytest.fixture(scope="function")
def example_subcomponent() -> Component:
    """
    Provides an example subcomponent using the generic Component class.
    """
    return AssonantComponentFactory.create_component_by_class_type(Component, "subcomponent")


@pytest.fixture(scope="function")
def example_components_collection() -> Dict[str, Component]:
    """
    Provides an example subcomponent using the generic Component class.
    """

    component_collection = {}

    for name, _ in inspect.getmembers(importlib.import_module("assonant.data_classes.components"), inspect.isclass):
        component_collection[name] = AssonantComponentFactory.create_component_by_class_name(name, name.lower())

    return component_collection


@pytest.fixture(scope="session")
def component_to_nexus_mapping() -> Dict[Component, Union[NXgroup, NXfield, NXroot]]:
    """
    Provides a dict that maps the Assonant Component to its respective NeXus object.
    """
    component_to_nexus_class_map = {
        Wiggler: NXinsertion_device,
        Undulator: NXinsertion_device,
        BendingMagnet: NXbending_magnet,
        Grating: NXgrating,
        Sensor: NXsensor,
        Detector: NXdetector,
        Monochromator: NXmonochromator,
        Mirror: NXmirror,
        Shutter: NXbeam_stop,
        Slit: NXslit,
        Collimator: NXcollimator,
        StorageRing: NXsource,
        Sample: NXsample,
        Entry: NXentry,
        BeamStopper: NXbeam_stop,
        MonochromatorCrystal: NXcrystal,
        MonochromatorVelocitySelector: NXvelocity_selector,
        BVS: NXdetector,
        Beamline: NXinstrument,
        Beam: NXbeam,
        Pinhole: NXpinhole,
        FresnelZonePlate: NXfresnel_zone_plate,
        DetectorROI: NXgroup,  # This is a contributed definition so current bypassed as NXgroup
        DetectorChannel: NXdetector_channel,
        DetectorModule: NXdetector_module,
        Cryojet: NXgroup,  # This is a custom group created so current bypassed as NXgroup
        GraniteBase: NXgroup,  # This is a custom group created so current bypassed as NXgroup
        Dewar: NXgroup,  # This is a custom group created so current bypassed as NXgroup
        Attenuator: NXattenuator,
        Entry: NXentry,
        Aperture: NXaperture,
    }

    return component_to_nexus_class_map


@pytest.fixture(scope="session")
def data_handler_to_nexus_mapping() -> Dict[DataHandler, Union[NXgroup, NXfield]]:
    """
    Provides a dict that maps the Assonant DataHandler to its respective NeXus object.
    """
    data_handler_to_nexus_class_map = {DataField: NXfield, Axis: [NXfield, NXlog], TimeSeries: NXlog}

    return data_handler_to_nexus_class_map



"""Tests focused on validating Hierarchizer methods"""

from unittest.mock import MagicMock

import pytest

from assonant.hierarchizer import Hierarchizer
from assonant.hierarchizer.exceptions import AssonantHierarchizerError


@pytest.fixture
def mock_metadata_retriever() -> MagicMock:
    """Fixture that returns a mocked AssonantMetadataRetriever."""
    mock = MagicMock()
    return mock


def test_init_with_custom_metadata_retriever(mock_metadata_retriever):
    """Ensure constructor uses provided metadata retriever instead of creating a new one.

    Expected Behavior:
        - Hierarchizer internal metadata_retriever must be equal to the one passed
        in its constructor method
    """
    hier = Hierarchizer("fake_schema.csv", metadata_retriever=mock_metadata_retriever)
    assert hier.metadata_retriever is mock_metadata_retriever


def test_init_with_invalid_schema_path_raises():
    """Ensure exception is wrapped in AssonantHierarchizerError if retriever fails.

    Expected Behavior:
        - Hierarchizer must raise and AssonantHierarchizerError to signalize invalid schema
    """
    with pytest.raises(AssonantHierarchizerError):
        Hierarchizer("invalid_schema.csv")


def test_mount_tree_node_set_root(mock_metadata_retriever):
    """Ensure mount_tree_node_set method correctly mount set for root Component (Component
    without parent).

    Expected Behavior:
        - If a component has no parent, it should return a set with only its tuple
    """
    hier = Hierarchizer("schema.csv", metadata_retriever=mock_metadata_retriever)
    root_tuple = ("root", "Beamline", None)
    result = hier._mount_tree_node_set(root_tuple)
    assert result == {root_tuple}


def test_mount_tree_node_set_with_parent(mock_metadata_retriever):
    """Ensure mount_tree_node_set method correctly mount set when component there are
    components with parent components.

    Expected Behavior:
        - If a component has a parent, the returned set must contain it and also its
        parent. This behavior must be recursive until gets to a root Component
    """
    hier = Hierarchizer("schema.csv", metadata_retriever=mock_metadata_retriever)

    parent_info = {"component_info": {"name": "root", "class": "Beamline"}}
    mock_metadata_retriever.get_component_info.return_value = parent_info

    child_tuple = ("child", "Mirror", "root")
    result = hier._mount_tree_node_set(child_tuple)

    assert result == {("child", "Mirror", "root"), ("root", "Beamline", None)}


def test_hierarchize_components_with_entry_and_beamline(mock_metadata_retriever, example_entry, example_beamline):
    """Ensure hierarchize_components method correctly handle special case of Entry
    and Beamline Components.

    Expected Behavior:
        - Entry and Beamline should remain unchanged and returned as-is as they are special cases in the hierarchy
        and must be ignored by Hierarchizer.
    """
    hier = Hierarchizer("schema.csv", metadata_retriever=mock_metadata_retriever)

    components = [example_entry, example_beamline]
    result = hier.hierarchize_components(components)

    # Should return unchanged
    assert result == components


def test_hierarchize_components_without_hierarchy(mock_metadata_retriever, example_components_collection):
    """Ensure hierarchy is built correctly for list of components without parent (root components).

    Expected Behavior:
        - The list returned by the hierarchizer must be equal to the input list.
    """
    hier = Hierarchizer("schema.csv", metadata_retriever=mock_metadata_retriever)

    monoch = example_components_collection["Monochromator"]
    mirror = example_components_collection["Mirror"]
    input_monoch = example_components_collection["Monochromator"].model_copy(deep=True)
    input_mirror = example_components_collection["Mirror"].model_copy(deep=True)

    # This is the expected result
    monoch.add_subcomponent(mirror)

    # Mock Metadata retriever to return that mirror is subcomponent of monochromator
    def get_info(name):
        """MetadataRetriever.get_component_info method mock"""
        if name == monoch.name:
            return {"component_info": {"name": "monochromator", "class": "Monochromator"}}
        elif name == mirror.name:
            return {"component_info": {"name": "mirror", "class": "Mirror"}}

    mock_metadata_retriever.get_component_info.side_effect = get_info

    components = [input_monoch, input_mirror]
    result = hier.hierarchize_components(components)

    # Iterate over result to compare lists. As order the same order is not guaranteed and Components are
    # not hashable to be converted to a set or sortable, manually compare if elements from a list
    # is within the other
    assert all(component in result for component in components)


def test_hierarchize_components_with_simple_hierarchy(mock_metadata_retriever, example_components_collection):
    """Ensure hierarchy is built correctly for components with parent-child relation.

    Expected Behavior:
        - The list returned by the hierarchizer must be different from input list.
        - Child components must be within its parent component subcomponents group
        - Returned list must contain only root Components (non-parent Components)
    """
    hier = Hierarchizer("schema.csv", metadata_retriever=mock_metadata_retriever)

    monoch = example_components_collection["Monochromator"]
    mirror = example_components_collection["Mirror"]
    input_monoch = example_components_collection["Monochromator"].model_copy(deep=True)
    input_mirror = example_components_collection["Mirror"].model_copy(deep=True)

    # This is the expected result
    monoch.add_subcomponent(mirror)

    # Mock Metadata retriever to return that mirror is subcomponent of monochromator
    def get_info(name):
        """MetadataRetriever.get_component_info method mock"""
        if name == monoch.name:
            return {"component_info": {"name": "monochromator", "class": "Monochromator"}}
        elif name == mirror.name:
            return {"component_info": {"name": "mirror", "class": "Mirror", "subcomponent_of": "monochromator"}}

    mock_metadata_retriever.get_component_info.side_effect = get_info

    components = [input_monoch, input_mirror]
    result = hier.hierarchize_components(components)

    assert result != components
    assert len(result) == 1
    assert result[0] == monoch


def test_hierarchize_components_with_multi_subcomponent_hierarchy(
    mock_metadata_retriever, example_components_collection
):
    """Ensure hierarchy is built correctly for components with many parent-child relation.

    Expected Behavior:
        - The list returned by the hierarchizer must be different from input list.
        - Child components must be within its parent component subcomponents group
        - Returned list must contain only root Components (non-parent Components) with its children
        components within it
    """
    hier = Hierarchizer("schema.csv", metadata_retriever=mock_metadata_retriever)

    monoch = example_components_collection["Monochromator"]
    mirror = example_components_collection["Mirror"]
    detector = example_components_collection["Detector"]
    attenuator = example_components_collection["Attenuator"]

    input_monoch = example_components_collection["Monochromator"].model_copy(deep=True)
    input_mirror = example_components_collection["Mirror"].model_copy(deep=True)
    input_detector = example_components_collection["Detector"].model_copy(deep=True)
    input_attenuator = example_components_collection["Attenuator"].model_copy(deep=True)

    # This is the expected result
    monoch.add_subcomponent(mirror)
    attenuator.add_subcomponent(detector)
    monoch.add_subcomponent(attenuator)

    # Mock Metadata retriever to return that mirror is subcomponent of monochromator
    def get_info(name):
        """MetadataRetriever.get_component_info method mock"""
        if name == monoch.name:
            return {"component_info": {"name": "monochromator", "class": "Monochromator"}}
        elif name == mirror.name:
            return {"component_info": {"name": "mirror", "class": "Mirror", "subcomponent_of": "monochromator"}}
        elif name == attenuator.name:
            return {"component_info": {"name": "attenuator", "class": "Attenuator", "subcomponent_of": "monochromator"}}
        elif name == detector.name:
            return {"component_info": {"name": "detector", "class": "Detector", "subcomponent_of": "attenuator"}}

    mock_metadata_retriever.get_component_info.side_effect = get_info

    components = [input_monoch, input_mirror, input_detector, input_attenuator]
    result = hier.hierarchize_components(components)

    assert result != components
    assert len(result) == 1
    assert result[0] == monoch


def test_hierarchize_components_creates_missing_components_in_hierarchy(
    mock_metadata_retriever, example_components_collection
):
    """Ensure hierarchy is built correctly when a child component is passed but its
    parent component is not in the input list.

    Expected Behavior:
        - The list returned by the hierarchizer must be different from input list.
        - The Child components must be within its respective parent component
        internal subcomponents group
        - The Non-passed parent Component must be created internally by hierarchizer, even
        if it is a root Component.
        - Returned list must contain only root Component (non-parent Components) with
        its children component within it
    """
    hier = Hierarchizer("schema.csv", metadata_retriever=mock_metadata_retriever)

    monoch = example_components_collection["Monochromator"]
    mirror = example_components_collection["Mirror"]
    input_mirror = example_components_collection["Mirror"].model_copy(deep=True)

    # This is the expected result
    monoch.add_subcomponent(mirror)

    # Mock Metadata retriever to return that mirror is subcomponent of monochromator
    def get_info(name):
        """MetadataRetriever.get_component_info method mock"""
        if name == monoch.name:
            return {"component_info": {"name": "monochromator", "class": "Monochromator"}}
        elif name == mirror.name:
            return {"component_info": {"name": "mirror", "class": "Mirror", "subcomponent_of": "monochromator"}}

    mock_metadata_retriever.get_component_info.side_effect = get_info

    # Only pass the subcomponent as input to validated that hierarchizer creates parent component
    components = [input_mirror]
    result = hier.hierarchize_components(components)

    assert result != components
    assert len(result) == 1
    assert result[0] == monoch


def test_hierarchize_components_with_multi_missing_subcomponent_hierarchy(
    mock_metadata_retriever, example_components_collection
):
    """Ensure hierarchy is built correctly when many child components are passed but their
    parent components are not in the input list.

    Expected Behavior:
        - The list returned by the hierarchizer must be different from input list.
        - All Child components must be within their respective parent component
        internal subcomponents group
        - All Non-passed parent Components must be created internally by hierarchizer, even
        if it is a root Component.
        - Returned list must contain only root Component (non-parent Components) with
        its children components within it
    """
    hier = Hierarchizer("schema.csv", metadata_retriever=mock_metadata_retriever)

    monoch = example_components_collection["Monochromator"]
    mirror = example_components_collection["Mirror"]
    detector = example_components_collection["Detector"]
    attenuator = example_components_collection["Attenuator"]

    input_mirror = example_components_collection["Mirror"].model_copy(deep=True)
    input_detector = example_components_collection["Detector"].model_copy(deep=True)

    # This is the expected result
    monoch.add_subcomponent(mirror)
    attenuator.add_subcomponent(detector)
    monoch.add_subcomponent(attenuator)

    # Mock Metadata retriever to return that mirror is subcomponent of monochromator
    def get_info(name):
        """MetadataRetriever.get_component_info method mock"""
        if name == monoch.name:
            return {"component_info": {"name": "monochromator", "class": "Monochromator"}}
        elif name == mirror.name:
            return {"component_info": {"name": "mirror", "class": "Mirror", "subcomponent_of": "monochromator"}}
        elif name == attenuator.name:
            return {"component_info": {"name": "attenuator", "class": "Attenuator", "subcomponent_of": "monochromator"}}
        elif name == detector.name:
            return {"component_info": {"name": "detector", "class": "Detector", "subcomponent_of": "attenuator"}}

    mock_metadata_retriever.get_component_info.side_effect = get_info

    components = [input_mirror, input_detector]
    result = hier.hierarchize_components(components)

    assert result != components
    assert len(result) == 1
    assert result[0] == monoch



"""
Tests focused on validating AssonantDataLogger methods

This tests must not verify correct behavior of underlying modules
as that is part of their domain-specific test. This must only
validate AssonantDataLogger does not break during the
the coordination of the usage of all its components.
"""

import json
import os

import numpy as np
import pytest
from nexusformat.nexus.tree import nxload

from assonant.data_classes.components import Mirror, Monochromator, MonochromatorCrystal
from assonant.data_classes.enums import TransformationType
from assonant.data_classes.factories import AssonantComponentFactory
from assonant.data_logger import AssonantDataLogger
from assonant.naming_standards import AcquisitionType, BeamlineName, ExperimentStage

from ..utils import assert_n_files_exists_in_dir, get_fixture_values


# --------------------
# Fixtures
# --------------------
@pytest.fixture(scope="session")
def data_logger_pre_init_auto_collect(example_csv_metadata_file_path):
    """
    Provides a AssonantDataLogger configures to access metadata from example csv file
    and with the pre_init_aut_collector_pv_connections config on.
    """
    logger = AssonantDataLogger(
        BeamlineName.CATERETE,
        data_schema_file_path=example_csv_metadata_file_path,
    )
    return logger


@pytest.fixture(scope="function")
def data_logger(example_csv_metadata_file_path):
    """
    Provides a AssonantDataLogger configures to access metadata from example csv file
    and with the pre_init_aut_collector_pv_connections config off.
    """
    logger = AssonantDataLogger(BeamlineName.CATERETE, data_schema_file_path=str(example_csv_metadata_file_path))
    return logger


@pytest.fixture(scope="session")
def example_collected_pvs():
    """
    Provides a simulated dict containing simulated data acquired for PVs from
    the example metadata csv.
    """
    return {
        "ATTENUATOR_PV": 10,
        "APERTURE_PV": "Test Repetition",
        "BEAM_PV": None,
        "BEAM STOPPER_PV": [1, 2, 3, 4, 5, 6],
        "COLLIMATOR_PV": 1.2,
        "MIRROR_PV": np.array([1, 2, None, 4, 5]),
        "THIS_PV_DONT_EXIST": "Test",
    }


@pytest.fixture(scope="session")
def example_collected_component_with_none_values():
    """
    Provides a simulated collected component with DataHandlers containing None value within it simulating
    may collection failure cases
    """
    mirror_with_none_values = AssonantComponentFactory.create_component_by_class_type(Mirror, "none_values_mirror")
    mirror_with_none_values.create_and_add_position(
        name="x", transformation_type=TransformationType.TRANSLATION, value=None, unit="cm"
    )
    mirror_with_none_values.create_and_add_position(
        name="y", transformation_type=TransformationType.TRANSLATION, value=[7.5, 8.0, 9.0]
    )
    mirror_with_none_values.create_and_add_position(
        name="Ry", transformation_type=TransformationType.ROTATION, value=[7.5, None, 9.0]
    )
    mirror_with_none_values.create_and_add_timeseries_position(
        name="z",
        transformation_type=TransformationType.TRANSLATION,
        value=[1, 2, 3, 4, 5],
        timestamps=[0.1, 0.2, None, 0.4, 0.5],
    )
    mirror_with_none_values.create_and_add_field(name="field1", value="some information")
    mirror_with_none_values.create_and_add_field(name="field2", value=["1", "2", None, "4", "5"], unit="GeV")

    return mirror_with_none_values


@pytest.fixture(scope="function")
def example_collected_component_with_subcomponent():
    """
    Provides a simulated collected component with subcomponent
    """
    monoch_crystal = AssonantComponentFactory.create_component_by_class_type(MonochromatorCrystal, "crystal_2")
    monoch_crystal.create_and_add_position(
        name="y", transformation_type=TransformationType.TRANSLATION, value=[7.5, 8.0, 9.0]
    )
    monoch_crystal.create_and_add_timeseries_position(
        name="z",
        transformation_type=TransformationType.TRANSLATION,
        value=[1, 2, 3, 4, 5],
        timestamps=[0.1, 0.2, 0.3, 0.4, 0.5],
    )
    monoch_crystal.create_and_add_position(
        name="Rx", transformation_type=TransformationType.ROTATION, value=20, unit="rad"
    )
    monoch_crystal.create_and_add_field(name="temperature", value=30, unit="K")

    monoch = AssonantComponentFactory.create_component_by_class_type(Monochromator, "mono_fields")
    monoch.add_subcomponent(monoch_crystal)
    monoch.create_and_add_field(name="wavelength", value=[10, 20, 30, 40], unit="db")
    monoch.create_and_add_field(
        name="energy",
        value=10,
        extra_metadata={
            "attr1": 10,
            "attr2": "atributo 2",
            "attr3": 0.765,
            "attr4": [1, 2, 3, 4, 5],
        },
    )

    return monoch


@pytest.fixture(scope="function")
def example_collected_component_without_subcomponent():
    mirror = Monochromator(name="mirror")
    mirror.create_and_add_field(name="field_1", value=[10, 20, 30, 40], unit="u")
    mirror.create_and_add_field(
        name="field_2",
        value=10,
        extra_metadata={
            "attr1": 10,
            "attr2": "atributo 2",
            "attr3": 0.765,
            "attr4": [1, 2, 3, 4, 5],
        },
    )

    return mirror


# --------------------
# Tests
# --------------------


def test_log_collected_pvs(data_logger, example_collected_pvs, tmp_path):
    """
    Validate that AssonantDataLogger works correctly and don't break when using
    its method to log collect PV data for existing AcquisitionTypes.

    PS: Content is not being validated only if module does not break!

    Expected Behavior:
        - AssonantDatalogger must create a new NeXus file into target file path with the given file name
        - A entry with the name of the ExperimentStage must exist in the file.
        - Inside the ExperimentStage Entry, there must be other entries for each existing
        AcquisitionType.
        - Within each AcquisitionType Entry there must be a standardized structure
        with the data related to each AcquisitionType
    """
    log_file_path = tmp_path
    log_file_name = "test_file.nxs"

    for acq_type in AcquisitionType:
        data_logger.log_collected_pv_data(
            log_file_path, log_file_name, ExperimentStage.SAMPLE_ACQUISITION, acq_type, example_collected_pvs
        )

    assert_n_files_exists_in_dir(1, log_file_path)
    assert os.path.exists(os.path.join(log_file_path, log_file_name))


def test_log_empty_pvs_dict(data_logger, tmp_path):
    """
    Validate that AssonantDataLogger works correctly and don't break when using
    its method to log collect PV data for existing AcquisitionTypes but data dict
    is empty.

    PS: Content is not being validated only if module does not break!

     Expected Behavior:
        - AssonantDatalogger must create a new NeXus file into target file path with the given file name
        - A entry with the name of the ExperimentStage must exist in the file.
        - Inside the ExperimentStage Entry, there must be other entries for each existing
        AcquisitionType.
        - Within each AcquisitionType Entry no data must exist.
    """
    log_file_path = tmp_path
    log_file_name = "test_file.nxs"

    for acq_type in AcquisitionType:
        data_logger.log_collected_pv_data(
            log_file_path, log_file_name, ExperimentStage.SAMPLE_ACQUISITION, acq_type, {}
        )


def test_log_component_list(
    data_logger,
    example_collected_component_with_subcomponent,
    example_collected_component_without_subcomponent,
    tmp_path,
):
    """
    Validate that AssonantDataLogger works correctly and don't break when using
    its method to log collect data represented within Assonant Componets wrapped
    in a list.

    PS: Content is not being validated only if module does not break!

    Expected Behavior:
        - AssonantDatalogger must create a new NeXus file into target file path with the given file name
        - A entry with the name of the ExperimentStage must exist in the file.
        - Inside the ExperimentStage Entry, there must be other entries for each existing
        AcquisitionType.
        - Within each AcquisitionType Entry there must be a standardized structure
        with the data related to each AcquisitionType
    """
    log_file_path = tmp_path
    log_file_name = "test_file.nxs"
    exp_stage = ExperimentStage.SAMPLE_ACQUISITION

    component_list = [example_collected_component_with_subcomponent, example_collected_component_without_subcomponent]

    for acq_type in AcquisitionType:
        data_logger.log_collected_component_data(log_file_path, log_file_name, exp_stage, acq_type, component_list)
    assert_n_files_exists_in_dir(1, log_file_path)
    assert os.path.exists(os.path.join(log_file_path, log_file_name))


def test_log_single_component(data_logger, example_collected_component_without_subcomponent, tmp_path):
    """
    Validate that AssonantDataLogger works correctly and don't break when using
    its method to log collect data represented within Assonant Componets and a
    single Component is passed.

    PS: Content is not being validated only if module does not break!

    Expected Behavior:
        - AssonantDatalogger must create a new NeXus file into target file path with the given file name
        - A entry with the name of the ExperimentStage must exist in the file.
        - Inside the ExperimentStage Entry, there must be other entries for each existing
        AcquisitionType.
        - Within each AcquisitionType Entry there must be a standardized structure
        with the data related to each AcquisitionType
    """
    log_file_path = tmp_path
    log_file_name = "test_file.nxs"
    exp_stage = ExperimentStage.SAMPLE_ACQUISITION

    for acq_type in AcquisitionType:
        data_logger.log_collected_component_data(
            log_file_path,
            log_file_name,
            exp_stage,
            acq_type,
            example_collected_component_without_subcomponent,
        )


@pytest.mark.parametrize(
    "bluesky_path_fixture_name",
    [
        "example_bluesky_baseline_acquisition_file_path",
        "example_bluesky_primary_acquisition_file_path",
        "example_bluesky_monitor_acquisition_file_path",
        "example_bluesky_flyscan_acquisition_file_path",
        "example_bluesky_all_streams_acquisition_file_path",
    ],
)
def test_log_bluesky_acquisition_from_json_file(data_logger, bluesky_path_fixture_name, tmp_path, request):
    """
    Validate that AssonantDataLogger works correctly and don't break when using
    its method to log bluesky collected data within passed JSON file.

    PS: Content is not being validated only if module does not break!

    Expected Behavior:
        - AssonantDatalogger must create a new NeXus file into target file path with the given file name
        - A entry with the name of the ExperimentStage must exist in the file.
        - Inside the ExperimentStage Entry, there must be other entries for each existing
        AcquisitionType.
        - Within each AcquisitionType Entry there must be a standardized structure
        with the data related to each AcquisitionType
    """
    log_file_path = tmp_path
    log_file_name = "test_file.nxs"
    exp_stage = ExperimentStage.SAMPLE_ACQUISITION

    # Get the fixture value dynamically
    example_bluesky_file_path = get_fixture_values(bluesky_path_fixture_name, request)

    # Run log for each acquisition type
    for acq_type in AcquisitionType:
        data_logger.log_bluesky_data_from_json_file(
            log_file_path,
            log_file_name,
            exp_stage,
            acq_type,
            example_bluesky_file_path,
        )

    # Assert that only one file was created
    files = list(tmp_path.iterdir())
    assert len(files) == 1
    assert files[0].name == log_file_name

    # Assert that 1 entry was created for each acquisition type
    nxroot = nxload(os.path.join(log_file_path, log_file_name), mode="r")
    assert len(nxroot[exp_stage.value].entries) == len(AcquisitionType)


@pytest.mark.parametrize(
    "bluesky_path_fixture_name",
    [
        "example_bluesky_baseline_acquisition_file_path",
        "example_bluesky_primary_acquisition_file_path",
        "example_bluesky_monitor_acquisition_file_path",
        "example_bluesky_flyscan_acquisition_file_path",
        "example_bluesky_all_streams_acquisition_file_path",
    ],
)
def test_log_bluesky_acquisition_from_docs_list(data_logger, bluesky_path_fixture_name, tmp_path, request):
    """
    Validate that AssonantDataLogger works correctly and don't break when using
    its method to log bluesky collected data within list of bluesky docs
    read from the JSON file.

    PS: Content is not being validated only if module does not break!

    Expected Behavior:
        - AssonantDatalogger must create a new NeXus file into target file path with the given file name
        - A entry with the name of the ExperimentStage must exist in the file.
        - Inside the ExperimentStage Entry, there must be other entries for each existing
        AcquisitionType.
        - Within each AcquisitionType Entry there must be a standardized structure
        with the data related to each AcquisitionType
    """
    log_file_path = tmp_path
    log_file_name = "test_file.nxs"
    exp_stage = ExperimentStage.SAMPLE_ACQUISITION

    # Get the fixture value dynamically
    example_bluesky_file_path = get_fixture_values(bluesky_path_fixture_name, request)
    with open(example_bluesky_file_path) as f:
        example_bluesky_docs_list = json.load(f)

    # Run log for each acquisition type
    for acq_type in AcquisitionType:
        data_logger.log_bluesky_data_from_documents_list(
            log_file_path,
            log_file_name,
            exp_stage,
            acq_type,
            example_bluesky_docs_list,
        )

    # Assert that only one file was created
    files = list(tmp_path.iterdir())
    assert len(files) == 1
    assert files[0].name == log_file_name

    # Assert that 1 entry was created for each acquisition type
    nxroot = nxload(os.path.join(log_file_path, log_file_name), mode="r")
    assert len(nxroot[exp_stage.value].entries) == len(AcquisitionType)



"""Tests focused on validating AssonantMetadataRetriever methods"""

import pytest

from assonant.metadata_retriever import AssonantMetadataRetriever
from assonant.metadata_retriever.enums import CSVColumn
from assonant.metadata_retriever.exceptions import MetadataRetrieverFactoryError

from ..utils import assert_structure_has_keys

# NOTE: Most of tests (if not all) are not validating returned content as
# CSVMetadataRetriever currently alter value formatting. That said just
# returned structure is being asserted on those tests. If a refactor of
# the CSVMetadataRetriever and AssonantMetadataRetriever occur, checking
# returned content may make sense.

# Define mandatory and optinal fields dict for get_component_info() method returned dict
GET_COMPONENT_INFO_MANDATORY_COMPONENT_KEYS = {"name", "class"}
GET_COMPONENT_INFO_OPTIONAL_COMPONENT_KEYS = {"subcomponent_of"}

# Define mandatory and optinal fields dict for get_pvs_info() method returned dict
GET_PV_INFO_MANDATORY_COMPONENT_KEYS = {"name", "class"}
GET_PV_INFO_OPTIONAL_COMPONENT_KEYS = {"subcomponent_of"}
GET_PV_INFO_MANDATORY_HANDLER_KEYS = {"name", "value", "pv_name"}
GET_PV_INFO_OPTIONAL_HANDLER_KEYS = {"unit", "transformation_type"}


@pytest.fixture(scope="session")
def example_csv_metadata_retriever(example_csv_metadata_file_path) -> AssonantMetadataRetriever:
    """
    Provides a functional AssonantMetadataRetriever for the example_metadata.csv.
    """
    return AssonantMetadataRetriever(metadata_source_file_path=example_csv_metadata_file_path)


def test_invalid_file_format():
    """
    Validate that AssonantMetadataRetriever constructor correctly raises exception
    if path to a file of an unsupported file format is passed during object construction.

    Expected Behavior:
        - AssonantMetadataRetriever must raise an MetadataRetrieverFactoryError if an filepath for unsupported
        file format is passed.
    """
    with pytest.raises(MetadataRetrieverFactoryError):
        _ = AssonantMetadataRetriever(metadata_source_file_path="UNSUPPORTED_FILE_FORMAT")


def test_get_component_info_returned_structure_for_existing_component(example_csv_metadata_retriever):
    """
    Validate that AssonantMetadataRetriever get_component_info() method correctly return expected
    value when component does exist.

    Expected Behavior:
        - AssonantMetadataRetriever be able to return a dictionary containing retrieved metadata
        from the component
        - The dictionary structure must follow exactly the one define at the
        IAssonantMetadataRetriever interface.
    """
    component_name = "ATTENUATOR"
    retrieved_metadata = example_csv_metadata_retriever.get_component_info(component_name)

    comp_info = retrieved_metadata["component_info"]
    assert_structure_has_keys(
        comp_info, GET_COMPONENT_INFO_MANDATORY_COMPONENT_KEYS, GET_COMPONENT_INFO_OPTIONAL_COMPONENT_KEYS
    )


def test_get_component_info_returned_structure_for_existing_subcomponent(example_csv_metadata_retriever):
    """
    Validate that AssonantMetadataRetriever get_component_info() method correctly return expected
    value when component does exist and is a subcomponent.

    Expected Behavior:
        - AssonantMetadataRetriever be able to return a dictionary containing retrieved metadata
        from the component
        - The dictionary structure must follow exactly the one define at the
        IAssonantMetadataRetriever interface.
        - The dictionary structure must contain the  subcomponent_of field
    """
    component_name = "SUBCOMP_SENSOR"
    retrieved_metadata = example_csv_metadata_retriever.get_component_info(component_name)

    comp_info = retrieved_metadata["component_info"]
    assert_structure_has_keys(
        comp_info, GET_COMPONENT_INFO_MANDATORY_COMPONENT_KEYS, GET_COMPONENT_INFO_OPTIONAL_COMPONENT_KEYS
    )


def test_get_component_info_for_non_existing_component(example_csv_metadata_retriever):
    """
    Validate that AssonantMetadataRetriever get_component_info() method correctly
    return expected value when component does not exist.

    Expected Behavior:
        - AssonantMetadataRetriever must not break or raise an exception in this case
        - The returned value must be None
    """
    retrived_metadata = example_csv_metadata_retriever.get_component_info("this component name does not exist")
    assert retrived_metadata == {}


def test_get_subcomponents_mapping(example_csv_metadata_retriever, example_metadata_csv_df):
    """
    Validate that AssonantMetadataRetriever get_subcomponents_mapping() method correctly
    retrieves a mapping of components that contains subcomponents.

    Expected Behavior:
        - AssonantMetadataRetriever must return the information in the expected structure defined
        in the IAssonantMetadataRetriever Interface
        - The dictionary structure must contain the correct number of components that contains subcomponents
        - component and subcomponents names must be correct
    """
    n_of_components_with_subcomponents = example_metadata_csv_df[CSVColumn.SUBCOMPONENT_OF.value].nunique(dropna=True)
    n_of_subcomponents = sum(example_metadata_csv_df[CSVColumn.SUBCOMPONENT_OF.value].value_counts(dropna=True))

    mapping = example_csv_metadata_retriever.get_subcomponents_mapping()

    assert len(mapping.keys()) == n_of_components_with_subcomponents
    assert n_of_subcomponents == sum([len(v) for v in mapping.values()])


def test_get_single_pv_info_return_mandatory_structure(example_csv_metadata_retriever):
    """
    Validate that AssonantMetadataRetriever get_pvs_info() method correctly
    retrieves info related to a PV when a single PV name is passed.

    Expected Behavior:
        - AssonantMetadataRetriever must return the information in the expected structure defined
        in the IAssonantMetadataRetriever Interface.
    """
    pv_name = "ATTENUATOR_PV"
    pv_info = example_csv_metadata_retriever.get_pvs_info(pv_name)

    assert pv_info[pv_name]

    comp_info = pv_info[pv_name]["component_info"]
    handler_info = pv_info[pv_name]["data_handler_info"]

    # Enforce contract
    assert_structure_has_keys(comp_info, GET_PV_INFO_MANDATORY_COMPONENT_KEYS, GET_PV_INFO_OPTIONAL_COMPONENT_KEYS)
    assert_structure_has_keys(handler_info, GET_PV_INFO_MANDATORY_HANDLER_KEYS, GET_PV_INFO_OPTIONAL_HANDLER_KEYS)


def test_get_multiple_pv_info(example_csv_metadata_retriever):
    """
    Validate that AssonantMetadataRetriever get_pvs_info() method correctly
    retrieves info related to when multiple PV names are passed as input.

    Expected Behavior:
        - AssonantMetadataRetriever must return the information in the expected structure defined
        in the IAssonantMetadataRetriever Interface.
    """
    pv_names = ["ATTENUATOR_PV", "SENSOR_PV"]
    pvs_info = example_csv_metadata_retriever.get_pvs_info(pv_names)

    assert all(pv_name in pvs_info.keys() for pv_name in pv_names)

    for pv_name in pv_names:
        comp_info = pvs_info[pv_name]["component_info"]
        handler_info = pvs_info[pv_name]["data_handler_info"]

        # Enforce contract
        assert_structure_has_keys(comp_info, GET_PV_INFO_MANDATORY_COMPONENT_KEYS, GET_PV_INFO_OPTIONAL_COMPONENT_KEYS)
        assert_structure_has_keys(handler_info, GET_PV_INFO_MANDATORY_HANDLER_KEYS, GET_PV_INFO_OPTIONAL_HANDLER_KEYS)


def test_get_unexisting_pv_info(example_csv_metadata_retriever):
    """
    Validate that AssonantMetadataRetriever get_pvs_info() method correctly
    retrieves info related to a PV when an single PV name is passed.

    Expected Behavior:
        - AssonantMetadataRetriever must return the information in the expected structure defined
        in the IAssonantMetadataRetriever Interface.
    """
    pv_name = "This pv does not exist"
    pv_info = example_csv_metadata_retriever.get_pvs_info(pv_name)
    assert pv_info == {}



"""Assonant file writer exceptions.

This submodule defines Exceptions used all over the file writer module and its submodules.
"""

from .factories_exceptions import FileWriterFactoryError
from .file_writer_exceptions import (
    AssonantFileWriterError,
    FileAlreadyExistsError,
    NeXusObjectAlreadyExistsError,
)

__all__ = [
    "AssonantFileWriterError",
    "FileAlreadyExistsError",
    "FileWriterFactoryError",
    "NeXusObjectAlreadyExistsError",
]



"""Assonant Component abstract class."""

from typing import Dict, Generator, Iterable, List, Optional, Type, Union

import numpy as np

from ..assonant_data_class import AssonantDataClass
from ..data_handlers import Axis, DataField, DataHandler, TimeSeries
from ..enums import TransformationType
from ..exceptions.data_classes_exceptions import (
    AxisCreationError,
    AxisInsertionError,
    DataHandlerInsertionError,
    DataHandlerTakeageError,
    FieldCreationError,
    FieldInsertionError,
    SubcomponentInsertionError,
)


class Component(AssonantDataClass):
    """Abstract class that creates the base common requirements to define an Assonant Component.

    Components are more generic definitions which may be composed by many subcomponents if more
    detailing in its composition is desired.
    """

    name: str
    subcomponents: Optional[Dict[str, "Component"]] = {}
    positions: Optional[Dict[str, Axis]] = {}
    fields: Optional[Dict[str, DataHandler]] = {}

    def add_subcomponent(self, new_component: Union["Component", List["Component"]]):
        """Add new subcomponent or list of new subcomponents to component.

        Args:
            component (Union[Component, List[Component]]): Component object or List of Components which will be add as
            subcomponent from called Component object.

        Raises:
            SubcomponentInsertionError: Failure to insert Subcomponent to subcomponents dict from Component.
        """
        try:
            if isinstance(new_component, List):
                for _new_component in new_component:
                    if _new_component.name not in self.subcomponents:
                        self.subcomponents[_new_component.name] = _new_component
                    else:
                        # Generate a new name with a index to avoid equal names
                        new_name = self._generate_new_valid_name(
                            names_list=self.subcomponents, name=_new_component.name
                        )
                        _new_component.name = new_name
                        self.subcomponents[new_name] = _new_component
            elif isinstance(new_component, Component):
                if new_component.name not in self.subcomponents:
                    self.subcomponents[new_component.name] = new_component
                else:
                    # Generate a new name with a index to avoid equal names
                    new_name = self._generate_new_valid_name(names_list=self.subcomponents, name=new_component.name)
                    new_component.name = new_name
                    self.subcomponents[new_name] = new_component
            else:
                raise SubcomponentInsertionError(f"Invalid type for component being add! ({type(new_component)})")
        except Exception as e:
            raise SubcomponentInsertionError(
                f"Failed to add {new_component} Subcomponent to {self.name} subcomponents dict."
            ) from e

    def add_position(self, name: str, new_axis: Axis):
        """Add new positional field to component.

        Args:
            name (str): Axis name.
            axis (Axis): Axis DataHandler object.

        Raises:
            AxisInsertionError: Failure to insert Axis to Component positions dict.
        """
        try:
            if name not in self.positions:
                self.positions[name] = new_axis
            else:
                # Generate a new name with a index to avoid equal names
                new_name = self._generate_new_valid_name(names_list=self.positions, name=name)
                self.positions[new_name] = new_axis
        except Exception:
            raise AxisInsertionError(f"Failed to add {name} Axis to {self.name} Component.")

    def create_and_add_position(
        self,
        name: str,
        transformation_type: TransformationType,
        value: Union[int, float, str, List, Type[np.ndarray], None],
        unit: Optional[str] = None,
        extra_metadata: Optional[Dict[str, Union[int, float, str, List, Type[np.ndarray], None]]] = {},
    ):
        """Create and add new positional field to component.

        Args:
            name (str): Axis name.
            transformation_type (TransformationType): Type of transformation done by axis.
            value (Union[int, float, str, List, Type[np.ndarray], None]): Value related to axis
            collected data.
            unit (Optional[str], optional): Measurement unit related to value parameter.
            Defaults to None.
            extra_metadata (Optional[Dict[str, Union[int, float, str, List, Type[np.ndarray],
            None]]], optional): Dictionary
            containing any aditional metadata related to collected data. Defaults to {}.

            Raises:
                AxisCreationError: Failure to create new Axis DataHandler with passed data.
                AxisInsertionError: Failure to insert new Axis DataHanlder into Component
                positions dict.
        """
        from ..factories import AssonantDataHandlerFactory

        try:
            new_axis = AssonantDataHandlerFactory.create_axis(
                transformation_type=transformation_type,
                value=value,
                unit=unit,
                extra_metadata=extra_metadata,
            )
        except Exception as e:
            raise AxisCreationError(f"Failed to create '{name}' Axis DataHandler for {self.name} Component.") from e
        try:
            self.add_position(name=name, new_axis=new_axis)
        except AxisInsertionError as e:
            raise e

    def create_and_add_timeseries_position(
        self,
        name: str,
        transformation_type: TransformationType,
        value: Union[int, float, str, List, Type[np.ndarray], None],
        timestamps: Union[int, float, str, List, Type[np.ndarray], None],
        unit: Optional[str] = None,
        extra_metadata: Optional[Dict[str, Union[int, float, str, List, Type[np.ndarray], None]]] = {},
        timestamps_unit: Optional[str] = None,
        timestamp_extra_metadata: Optional[Dict[str, Union[int, float, str, List, Type[np.ndarray], None]]] = {},
    ):
        """Create and Add new positional field to component.

        Args:
            name (str): Axis name
            transformation_type (TransformationType): Transformation type of the Axis
            value (Union[int, float, str, List, Type[np.ndarray], None]): Value related to
            axis collected data
            timestamps (Union[int, float, str, List, Type[np.ndarray], None]): Timestamps
            related to data collected from the axis.
            unit (Optional[str], optional): Measurement unit related to value
            field. Defaults to None.
            extra_metadata (Optional[Dict[str, Union[int, float, str, List, Type[np.ndarray],
            None]]], optional): Dictionary
            containing extra metadata about value field. Defaults to {}.
            tracked it as a TimeSeries. Defaults to None.
            timestamps_unit (Optional[str], optional): Measurement unit related to
            timestamp field. Defaults to None.
            timestamp_extra_metadata (Optional[Dict[str, Union[int, float, str, List, Type[np.ndarray],
            None]]], optional): Dict
            containing extra metadata about timestamps field. Defaults to {}.

            Raises:
                AxisCreationError: Failure to create new Axis DataHandler with passed data.
                AxisInsertionError: Failure to insert new Axis DataHanlder into Component
                positions dict.
        """
        from ..factories import AssonantDataHandlerFactory

        try:
            new_axis = AssonantDataHandlerFactory.create_timeseries_axis(
                transformation_type=transformation_type,
                value=value,
                unit=unit,
                extra_metadata=extra_metadata,
                timestamps=timestamps,
                timestamps_unit=timestamps_unit,
                timestamp_extra_metadata=timestamp_extra_metadata,
            )
        except Exception as e:
            raise AxisCreationError(
                f"Failed to create '{name}' Axis Data TimeSeries with TimeSeries data for {self.name} Component."
            ) from e
        try:
            self.add_position(name=name, new_axis=new_axis)
        except AxisInsertionError as e:
            raise e

    def add_field(self, name: str, new_field: DataField):
        """Add new data field to component.

        Args:
            name (str): Field name
            data_field (DataField): DataField DataHandler object.

        Raises:
            FieldInsertionError: Failure on inserting DataField to Component fields dict.
        """
        try:
            if name not in self.fields:
                self.fields[name] = new_field
            else:
                # Generate a new name with a index to avoid equal names
                new_name = self._generate_new_valid_name(names_list=self.fields, name=name)
                self.fields[new_name] = new_field
        except Exception:
            raise FieldInsertionError(f"Failed to add '{name}' field to {self.name} Component.")

    def create_and_add_field(
        self,
        name: str,
        value: Union[int, float, str, List, Type[np.ndarray], None],
        unit: Optional[str] = None,
        extra_metadata: Optional[Dict[str, Union[int, float, str, List, Type[np.ndarray], None]]] = {},
    ):
        """Create and add new data field.

        Args:
            name (str): Field name.
            value (Union[int, float, str, List, Type[np.ndarray], None]): Value related to field
            collected data.
            unit (Optional[str], optional): Measurement unit related to value parameter.
            Defaults to None.
            extra_metadata (Optional[Dict[str, Union[int, float, str, List, Type[np.ndarray], None]]], optional): Dict
            containing any aditional metadata related to collected data. Defaults to {}.

        Raises:
            FieldCreationError: Failure on creating new DataField.
            FieldInsertionError: Failure on inserting DataField into Component fields dict.
        """
        from ..factories import AssonantDataHandlerFactory

        try:
            new_field = AssonantDataHandlerFactory.create_data_field(
                value=value,
                unit=unit,
                extra_metadata=extra_metadata,
            )
        except Exception as e:
            raise FieldCreationError(
                f"Failed to create '{name}' DataField Data Handler for {self.name} Component."
            ) from e
        try:
            self.add_field(name=name, new_field=new_field)
        except FieldInsertionError as e:
            raise e

    def add_timeseries_field(self, name: str, new_time_series_field: TimeSeries):
        """Add new data field to component that was collected as a TimeSeries.

        Args:
            name (str): Field name
            new_time_series_field (TimeSeries): TimeSeries DataHandler object.

        Raises:
            FieldInsertionError: Failure to insert TimeSeries to Component fields dict.
        """
        try:
            if name not in self.fields:
                self.fields[name] = new_time_series_field
            else:
                # Generate a new name with a index to avoid equal names
                new_name = self._generate_new_valid_name(names_list=self.fields, name=name)
                self.fields[new_name] = new_time_series_field
        except Exception:
            raise FieldInsertionError(f"Failed to add '{name}' TimeSeries field to {self.name} Component.")

    def create_and_add_timeseries_field(
        self,
        name: str,
        value: Union[int, float, str, List, Type[np.ndarray], None],
        timestamps: Union[int, float, str, List, Type[np.ndarray], None],
        unit: Optional[str] = None,
        extra_metadata: Optional[Dict[str, Union[int, float, str, List, Type[np.ndarray], None]]] = {},
        timestamps_unit: Optional[str] = None,
        timestamp_extra_metadata: Optional[Dict[str, Union[int, float, str, List, Type[np.ndarray], None]]] = {},
    ):
        """Create and add new data field to component that data was collected as a TimeSeries.

        Args:
            name (str): Field name.
            value (Union[int, float, str, List, Type[np.ndarray], None]): Value related to field
            collected data.
            timestamps (Union[int, float, str, List, Type[np.ndarray], None]): Timestamps related to data collected
            from the field.
            unit (Optional[str], optional): Measurement unit related to value parameter.
            Defaults to None.
            extra_metadata (Optional[Dict[str, Union[int, float, str, List, Type[np.ndarray],
            None]]], optional): Dictionary
            containing any aditional metadata related to collected data. Defaults to {}.
            timestamps_unit (Optional[str], optional): Measurement unit related to
            timestamp field. Defaults to None.
            timestamp_extra_metadata (Optional[Dict[str, Union[int, float, str, List, Type[np.ndarray],
            None]]], optional): Dictionary
            containing extra metadata about timestamps field. Defaults to {}.

            Raises:
                FieldCreationError: Failure on creating new DataField.
                FieldInsertionError: Failure on inserting DataField into Component fields dict.
        """
        from ..factories import AssonantDataHandlerFactory

        try:
            new_field = AssonantDataHandlerFactory.create_timeseries_field(
                value=value,
                unit=unit,
                extra_metadata=extra_metadata,
                timestamps=timestamps,
                timestamps_unit=timestamps_unit,
                timestamp_extra_metadata=timestamp_extra_metadata,
            )
        except Exception as e:
            raise FieldCreationError(
                f"Failed to create '{name}' TimeSeries Data Handler for {self.name} Component."
            ) from e
        try:
            self.add_timeseries_field(name=name, new_time_series_field=new_field)
        except FieldInsertionError as e:
            raise e

    def add_data_handler(self, name: str, new_data_handler: DataHandler):
        """Add new DataHandler to its specific dictionary within Component class.

        This method is a generic way to add any type of DataHandler to a component. It just
        call the specific underlying method to add that type of DataHandler to the Component.

        Args:
            name (str): Name attributed to passed DataHandler.
            new_data_handler (DataHandler): Any valid DataHandler object.

        Raises:
            DataHandlerInsertionError: Raised when the DataHandler type is not supported or an error occurs
            inside the specific method to add the DataHandlers to the component based in its type.
        """
        try:
            if isinstance(new_data_handler, DataField):
                self.add_field(name=name, new_field=new_data_handler)
            elif isinstance(new_data_handler, Axis):
                self.add_position(name=name, new_axis=new_data_handler)
            elif isinstance(new_data_handler, TimeSeries):
                self.add_timeseries_field(name=name, new_time_series_field=new_data_handler)
            else:
                raise DataHandlerInsertionError(
                    f"Adding DataHandler of type: {type(new_data_handler)} is not supported yet."
                )
        except Exception as e:
            raise DataHandlerInsertionError(
                f"Failed to add DataHandler of type {type(new_data_handler)} to {self.name} Component."
            ) from e

    def take_data_handlers_from(self, component: Union["Component", List["Component"]]):
        """Take DataHandlers from within a Component or list of Components and add to itself.

        PS: Taken DataHandlers are moved from target Components (passed component) to
        taker Component (Component which method was called from).

        Args:
            component (Union[Component, List[Component]]): Component or List of Components which DataHandlers
            will be taken.

        Raises:
            DataHandlerTakeageError: Raised when an error occurs when taking a DataHandler and trying to add
            to itself.
        """

        if not isinstance(component, List):
            components_list = [component]
        else:
            components_list = component

        for target_component in components_list:

            target_component_field_names = [*target_component.fields.keys()]
            target_component_position_names = [*target_component.positions.keys()]

            for field_name in target_component_field_names:
                try:
                    field_data_handler = target_component.fields.pop(field_name)
                    self.add_data_handler(name=field_name, new_data_handler=field_data_handler)
                except Exception as e:
                    raise DataHandlerTakeageError(
                        f"Failed when trying to take '{field_name}' DataHandler from '{target_component.name}' Component to add to '{self.name}' Component"
                    ) from e

            for position_name in target_component_position_names:
                try:
                    position_data_handler = target_component.positions.pop(position_name)
                    self.add_data_handler(name=position_name, new_data_handler=position_data_handler)
                except Exception as e:
                    raise DataHandlerTakeageError(
                        f"Failed when trying to take '{position_name}' DataHandler from '{target_component.name}' Component to add to '{self.name}' Component"
                    ) from e

    def _generate_new_valid_name(self, names_list: Iterable, name: str) -> str:
        """Generate a new name that don't exist on passed names list.

        Args:
            names_list (_type_): Iterable containing already existing names.
            name (str): Current name that will be changed to fit passed list.

        Returns:
            str: New unique name based on the passed name that don't exist on passed names list.
        """
        # Generate a new name with a index to avoid equal names
        new_name_generator = self._indexed_name_generator(name)
        new_name = next(new_name_generator)
        while new_name in names_list:
            new_name = next(new_name_generator)
        return new_name

    def _indexed_name_generator(self, name: str) -> Generator[str, None, None]:
        """Generator for creating indexed names.

        Args:
            name (str): Reference name to be indexed by generator.

        Yields:
            Generator[str, None, None]: New indexed name based on passed name.
        """
        i = 1
        while True:
            new_name = f"{name} ({i})"
            yield new_name
            i += 1



"""Exceptions triggered to warn about internal errors in data classes module."""


class AssonantDataClassesError(Exception):
    """Custom Exception to warn when any generic error occurs inside Assonant Data Classes Modules."""


class DataHandlerInsertionError(AssonantDataClassesError):
    """Custom Exception to warn when any error related to inserting data handler into Assonant Data Classes."""


class DataHandlerTakeageError(AssonantDataClassesError):
    """Custom Exception to warn when any error related to taking data handlers from another Assonant Data Classes."""


class FieldCreationError(AssonantDataClassesError):
    """Custom Exception to warn when any error related to creating fields for Assonant Data Classes."""


class FieldInsertionError(AssonantDataClassesError):
    """Custom Exception to warn when any error related to inserting fields into Assonant Data Classes."""


class AxisCreationError(AssonantDataClassesError):
    """Custom Exception to warn when any error related to creating positions for Assonant Data Classes."""


class AxisInsertionError(AssonantDataClassesError):
    """Custom Exception to warn when any error related to inserting positions into Assonant Data Classes."""


class SubcomponentInsertionError(AssonantDataClassesError):
    """Custom Exception to warn when any error related to inserting subcomponents into Assonant Data Classes."""



"""Assonant data retriever enums.

This submodule defines Enumerations classes used to standardize options related to data retrieving.
"""

from .data_source_file_format import DataSourceFileFormat

__all__ = ["AcquisitionType", "DataSourceFileFormat"]



"""Assonant Entry data class."""

from .component import Component


class Entry(Component):
    """Data classes that wraps data into a logical/temporal scope related to the experiment.

    Entries are used to group and represent data in a defined temporal/logical scope of the
    experiment.
    """



"""Exceptions triggered to warn about internal errors related to data classes Hierarchizer classes."""


class AssonantHierarchizerError(Exception):
    """Custom Exception to warn when any error occurs inside assonant data classes Hierarchizer class."""
